#ifndef DATASET_STUFF_H
#define DATASET_STUFF_H

#include <opencv2/opencv.hpp>

#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <vector>

/*****************************************************************/
struct Action{
  int startFrame;
  int endFrame;
  std::string verb;
  std::vector<std::string> names;
  void removeBadChars();
  void makeAllCharsSmall();
  void print() const;
  std::string getActionName() const;
};

/*****************************************************************/
std::vector<Action> readActionAnnotations_SMI(const char * filename, const int fps);

/*****************************************************************/
bool areEqualActions(Action a1, Action a2);
bool areEqualActions_verb(Action a1, Action a2);
struct Action_classcomp{
  bool operator() (const Action& a1, const Action& a2) const;
};

/*****************************************************************/

/*****************************************************************/
class DatasetSeq{
 public:
  DatasetSeq(){}
  DatasetSeq(const std::string& seqName, const int seqLength){
    seqName_ = seqName;
    seqLength_ = seqLength;
    startFrame_ = 1;
    endFrame_ = seqLength;
  }
  DatasetSeq(const std::string& seqName, const int seqLength, const int startFrame, const int endFrame){
    seqName_ = seqName;
    seqLength_ = seqLength;
    startFrame_ = startFrame;
    endFrame_ = endFrame;
  }
  int getSeqLength(){return seqLength_;}
  int getStartFrame(){return startFrame_;}
  int getEndFrame(){return endFrame_;}
  std::string getSeqName(){return seqName_;}
 private:
  int seqLength_;
  int startFrame_;
  int endFrame_;
  std::string seqName_;
};

/*****************************************************************/
class Dataset{
 public:  
  Dataset(const std::string& dbname, const std::string& dataDir, const float scale, const float resized, const int fps, const std::string imType){
    dbname_ = dbname;
    dataDir_ = dataDir;
    scale_ = scale;
    resized_ = resized;
    fps_ = fps;
    imType_ = imType;
  }
  float getScale() const{return scale_;}
  void setScale(const float scale){scale_ = scale;}
  float getResized() const{return resized_;}
  void setResized(const float resized){resized_ = resized;}
  int getFps() const{return fps_;}
  std::string getImType() const{return imType_;}
  DatasetSeq getSequence(const int seqNumber) {return sequences_[seqNumber];}
  void addSequence(const int seqNumber, const DatasetSeq seq){sequences_[seqNumber] = seq;}
  std::string getDatasetName() const{return dbname_;}
  std::string getDataDir() const {return dataDir_;}
  void setTrainSeqs(int version, std::vector<int> seqs){trainSeqs_[version] = seqs;}
  void setTestSeqs(int version, std::vector<int> seqs){testSeqs_[version] = seqs;}
  std::vector<int> getTrainSeqs(const int version){return trainSeqs_[version];}
  std::vector<int> getTestSeqs(const int version){return testSeqs_[version];}

  cv::Mat getIM(const int seqNumber, const int frameNumber);
  cv::Mat getGTObjectMask(const std::string& objectName, const int seqNumber, const int frameNumber);
  void writeGTObjectMask(const std::string& objectName, const int seqNumber, const int frameNumber, const cv::Mat& mask); 

  cv::Mat readUCM(const int seqNumber, const int frameNumber);

  void pushObjectName(std::string objectName){objectNames_.push_back(objectName);}
  std::vector<std::string> getObjectNames(){return objectNames_;}
  cv::Scalar getObjectColor(const std::string& objName);
  cv::Mat getObjectColors(const std::vector<std::string>& objNames);
  
  std::vector<Action> getActions(const int seqNumber);

  std::vector<Action> frequentActionsInTraining(const int trainVersion, const int minFreq);
  std::vector<Action> frequentActionsInTesting(const int trainVersion, const int minFreq);
  std::vector<Action> frequentActions(const int trainversion, const int minFreqTraining, const int minFreqTesting);
  int getActionFreqInTraining(const Action action, const int trainVersion);
  int getActionFreqInTesting(const Action action, const int trainVersion);

  std::vector<Action> getSeqFreqActions(const int seqNumber, const std::vector<Action> freqActions_);

  std::vector<std::pair<float,float> > getGazeLocations(const int seqNumber);
  std::pair<float,float> getGazeLocation(const int seqNumber, const int frame);
  std::string getGTObjectMaskPath(const std::string& objectName, const int seqNumber, const int frameNumber);

  cv::Mat readForegroundMask(const int seqNumber, const int frame, const int trainVersion);
  

 private:
  std::string getContourPath(const int seqNumber, const int frameNumber);
  std::string getImagePath(const int seqNumber, const int frameNumber);
  std::string getGazePath(const int seqNumber);
  std::string getFGSegmentationPath(const int seqNumber, const int frameNumber, const int trainVersion);

  std::string dbname_;
  std::string dataDir_;
  std::map<int,DatasetSeq> sequences_;
  float scale_;
  float resized_;
  int fps_;
  std::string imType_;
  std::map<int, std::vector<int> > trainSeqs_;
  std::map<int, std::vector<int> > testSeqs_;
  std::vector<std::string> objectNames_;
  std::map<int,std::vector<Action> > actions_;
  std::map<int,std::vector<std::pair<float,float> > > gazeLocations_;
};

/*****************************************************************/
Dataset dataset_setup(const std::string& datasetName);

/*****************************************************************/
Dataset dataset_setup_GTDA_v3();
/*****************************************************************/
Dataset dataset_setup_GTDA_v4();
/*****************************************************************/


void writeGTActions(Action action, const std::string& datasetName, const int trainVersion, std::vector<int> sequences, const bool doFG);


#endif //DATASET_STUFF_H
