#include "FeatureStuff.h"
#include <opencv2/gpu/gpu.hpp>

using namespace std;
using namespace cv;
using namespace cv::gpu;

/************************************************************************/
Mat getColorHistogram(const Mat & colorImage, const int binSize){
  int numBins = (int)(ceil(256.0/((double)(binSize))));
  //cout << "NumBins: " << numBins << endl;
  int height = colorImage.rows;
  int width = colorImage.cols;
  int ch = colorImage.channels();
  Mat output(height, width, CV_32S);

  typedef Vec<uchar,3> Vec3u;
  for(int i = 0; i < height; i++){
    for(int j = 0; j < width; j++){
      int b = colorImage.at<Vec3u>(i,j)[0];
      int g = colorImage.at<Vec3u>(i,j)[1];
      int r = colorImage.at<Vec3u>(i,j)[2];
      output.at<int>(i,j) = numBins*numBins*(b/binSize) + numBins*(g/binSize) + (r/binSize);
    }
  }

  int dummyY, dummyX;
  int alaki = matrixMaxValue<int>(output, dummyY, dummyX);
  //cout << "Maximum color histogram index: " << alaki << endl;

  return output;
}

/************************************************************************/
vector< vector<float> > getFeatureVectors_colorHist(const Mat & inputImage, const int binSize, const Mat& segmentation){
  int dummyY, dummyX;
  int m = matrixMaxValue<int>(segmentation, dummyY, dummyX);

  int numBins = (int)(ceil(256.0/((double)(binSize))));
  int histDim = numBins*numBins*numBins;

  Mat tempOutput(m, histDim, CV_32F);
  for(int i = 0; i < m; i++)
    for(int j = 0; j < histDim; j++)
      tempOutput.at<float>(i,j) = 0;

  Mat colorHist = getColorHistogram(inputImage, binSize);
  int height = colorHist.rows;
  int width = colorHist.cols;

  for(int i = 0; i < height; i++){
    for(int j = 0; j < width; j++){
      int s = segmentation.at<int>(i,j);
      int temp = colorHist.at<int>(i,j);
      tempOutput.at<float>(s-1,temp)++;
    }
  }

  vector< vector<float> > output;
  vector<float> tempVector;
  for(int j = 0; j < histDim; j++)
    tempVector.push_back(0);

  for(int i = 0; i < m; i++){
    double totalSum = 0;
    for(int j = 0; j < histDim; j++)
      totalSum += tempOutput.at<float>(i,j);
    for(int j = 0; j < histDim; j++)
      tempVector[j] = tempOutput.at<float>(i,j) / totalSum;
    output.push_back(tempVector);
  }
  
  return output;
}

/************************************************************************/
SparseMat getSpatial_colorHist(const vector< vector<float> > & fv_colorHist, const SparseMat & adjacencyMatrix, float var1, float var2){
  int m = (int)(fv_colorHist.size());
  
  int dims = adjacencyMatrix.dims();
  const int * sizes = adjacencyMatrix.size();
  if(m != sizes[0]){
    cout << "getSpatial_colorHist: adjacencyMatrix and colorHist features are not coming from the same segmentation" << endl;
    exit(1);
  }
  
  SparseMat output(dims, sizes, CV_32F);

  int idx[dims];
  for(SparseMat::const_iterator it = adjacencyMatrix.begin(); it != adjacencyMatrix.end(); it++){
    idx[0] = it.node()->idx[0];
    idx[1] = it.node()->idx[1];
    float temp = intersectHist(fv_colorHist[idx[0]], fv_colorHist[idx[1]]);
    temp = 1 - temp;
    output.ref<float>(idx) = exp((-pow(temp,var1)/var2));
  }

  return output;
}

/************************************************************************/
SparseMat getSpatialWeighted(const SparseMat & spatialMatrix, const SparseMat & adjacencyWeighted){
  int dims = spatialMatrix.dims();
  const int * sizes = spatialMatrix.size();
  int type = spatialMatrix.type();
  SparseMat output(dims, sizes, type);

  int idx[2];
  for(SparseMat::const_iterator it = spatialMatrix.begin(); it != spatialMatrix.end(); it++){
    idx[0] = it.node()->idx[0];
    idx[1] = it.node()->idx[1];
    output.ref<float>(idx) = spatialMatrix.value<float>(idx) * adjacencyWeighted.value<float>(idx);
  }

  return output;
}

/************************************************************************/
Mat getHOG(const Mat& im){
  //cout << "Computing HOG...";
  //cout.flush();

  int h = im.rows;
  int w = im.cols;

  Mat grayIm(h, w, CV_8UC1);
  cvtColor(im, grayIm, CV_BGR2GRAY, 1);

  cv::gpu::HOGDescriptor hog_desc(Size(16,16));
  GpuMat fv_gpu, im_gpu;
  im_gpu.upload(grayIm);
  hog_desc.getDescriptors(im_gpu, Size(8,8), fv_gpu);
  Mat fv;
  fv_gpu.download(fv);
  //cout << "getHOG " << fv.rows << " " << fv.cols << endl;

  //cout << "Done." << endl;

  return fv;
}

/************************************************************************/
cv::Mat getFeatureMap(const cv::Mat& fvs, const cv::Mat& centers){
  GpuMat fvs_gpu;
  GpuMat centers_gpu;
  fvs_gpu.upload(fvs);
  centers_gpu.upload(centers);
  vector<DMatch> matches;
  BruteForceMatcher_GPU<L2<float> > matcher;
  matcher.match(fvs_gpu, centers_gpu, matches);

  if(matches.size() != fvs.rows){
    cout << "getFeatureMap: Not all features are matched!" << endl;
    exit(1);
  }

  Mat output(fvs.rows, 1, CV_32S);
  for(int i = 0; i < fvs.rows; i++)
    output.at<int>(i,0) = matches[i].trainIdx;
  output = output + 1;

  return output;
}

/************************************************************************/
Mat featureMapHOG_reshape(const Mat& features, const int imrows, const int imcols){
  int tf = features.rows;
  int imrows_ = (imrows-8)/8;
  int imcols_ = (imcols-8)/8;
  int tf_ = imrows_ * imcols_;
  if(tf != tf_){
    cout << "getHOG_feature: rows and columns of image do not match the size of features!" << endl;
    exit(1);
  }
  
  Mat temp = features.reshape(0, imrows_);
  Mat output(imrows, imcols, CV_32S);
  resize(temp, output, Size(imcols, imrows), 0, 0, INTER_NEAREST);
  return output;
}

/************************************************************************/
Mat getColorLab(const Mat& im){
  int h = im.rows;
  int w = im.cols;
  typedef Vec<float,3> Vec3f;
  typedef Vec<uchar,3> Vec3u;

  Mat rgb_float(h, w, CV_32FC3);
  for(int i = 0; i < h; i++){
    for(int j = 0; j < w; j++){
      rgb_float.at<Vec3f>(i,j)[0] = (float)(im.at<Vec3u>(i,j)[0]) / 255.0;
      rgb_float.at<Vec3f>(i,j)[1] = (float)(im.at<Vec3u>(i,j)[1]) / 255.0;
      rgb_float.at<Vec3f>(i,j)[2] = (float)(im.at<Vec3u>(i,j)[2]) / 255.0;      
    }
  }  
  Mat Lab(h, w, CV_32FC3);
  cvtColor(rgb_float, Lab, CV_BGR2Lab);
  Mat output(h*w, 3, CV_32F);
  int count = 0;
  for(int i = 0; i < h; i++){
    for(int j = 0; j < w; j++){
      output.at<float>(count,0) = Lab.at<Vec3f>(i,j)[0];
      output.at<float>(count,1) = Lab.at<Vec3f>(i,j)[1];
      output.at<float>(count,2) = Lab.at<Vec3f>(i,j)[2];
      count++;
    }
  }
  return output;
}

/************************************************************************/
Mat featureMapLab_reshape(const Mat& features, const int imrows, const int imcols){
  int tf = features.rows;
  int tf_ = imrows * imcols;
  if(tf != tf_){
    cout << "getLab_feature: " << tf << "-" << tf_ << " rows and columns of image do not match the size of features!" << endl;
    exit(1);
  }
  
  Mat output = features.reshape(0, imrows);
  return output;
}


/************************************************************************/
Mat featureMap2regionFeatures(const Mat& featureMap, const Mat& segmentation, const int numBins){
  int dummyY, dummyX;
  int m = matrixMaxValue<int>(segmentation, dummyY, dummyX);

  int h = segmentation.rows;
  int w = segmentation.cols;

  Mat output = Mat::zeros(m, numBins, CV_32F);
  Mat sumput = Mat::zeros(m, 1, CV_32F);
  int s, f;
  for(int i = 0; i < h; i++){
    for(int j = 0; j < w; j++){
      s = segmentation.at<int>(i,j);
      f = featureMap.at<int>(i,j);
      output.at<float>(s-1, f-1)++;
      sumput.at<float>(s-1, 0)++;
    }
  }

  float myepsilon = 0.00001;

  for(int i = 0; i < m; i++)
    for(int j = 0; j < numBins; j++)
      output.at<float>(i,j) /= (sumput.at<float>(i,0) + myepsilon);

  return output;
}

/************************************************************************/
Mat getHierarchicalFeature(const Mat& fv2, const Mat& centers1, const Mat& segmentation2){
  int m1 = centers1.rows;
  int n = fv2.cols;
  Mat output(m1, n, CV_32F);
  int s, y, x;
  for(int i = 0; i < m1; i++){
    y = (int)(centers1.at<float>(i,0));
    x = (int)(centers1.at<float>(i,1));
    s = segmentation2.at<int>(y,x);
    for(int j = 0; j < n; j++)
      output.at<float>(i,j) = fv2.at<float>(s-1,j);
  }

  return output;
}

/************************************************************************/
Mat featureAroundPoint(const cv::Mat& featureMap, const float y, const float x, const int radius, const int numBins){
  Mat output = Mat::zeros(1, numBins, CV_32F);
  int cy = (int)(y);
  int cx = (int)(x);
  int ystart = max(cy-radius, 0);
  int yend = min(cy+radius, featureMap.rows-1);
  int xstart = max(cx-radius, 0);
  int xend = min(cx+radius, featureMap.cols-1);
  int dist;
  for(int i = ystart; i <= yend; i++){
    for(int j = xstart; j < xend; j++){
      dist = sqrt(pow((float)(i-cy), 2.0) + pow((float)(j-cx), 2.0));
      if(dist < radius)
	output.at<float>(0, featureMap.at<int>(i,j)-1)++;
    }
  }

  float sum = 0;
  for(int i = 0; i < numBins; i++)
    sum += output.at<float>(0,i);

  for(int i = 0; i < numBins; i++)
    output.at<float>(0,i) /= sum;

  return output;
}

/************************************************************************/
Mat featureInMask(const Mat& featureMap, const Mat& mask, const int numBins){
  Mat output = Mat::zeros(1, numBins, CV_32F);
  if(mask.type() != CV_8UC1){
    cout << "FeatureStuff: featureInMask: mask is not char type" << endl;
    return output;
  }

  for(int i = 0; i < featureMap.rows; i++)
    for(int j = 0; j < featureMap.cols; j++)
      if(mask.at<uchar>(i,j) >= 1){
	//cout << i << " " << j << " " << (int)(mask.at<uchar>(i,j)) << " " << featureMap.at<int>(i,j)-1 << endl;
	output.at<float>(0, featureMap.at<int>(i,j)-1)++;
      }


  float sum = 0;
  for(int i = 0; i < numBins; i++)
    sum += output.at<float>(0,i);

  if(sum == 0)
    return output;

  for(int i = 0; i < numBins; i++)
    output.at<float>(0,i) /= sum;

  return output;
}

/************************************************************************/
/************************************************************************/



