#include "MotionStuff.h"
#include <opencv2/gpu/gpu.hpp>


using namespace std;
using namespace cv;
using namespace cv::gpu;

/************************************************************************/
pair<Mat,Mat> getDenseFlow_eccv04(const Mat& image1, const Mat& image2, double resizeFactor){
  const float alpha_ = 0.075; //0.12
  const float gamma_ = 5; //5
  const float scale_factor_ = 0.95;
  const int inner_iterations_ = 5; //3
  const int outer_iterations_ = 100; //50
  const int solver_iterations_ = 30; //20

  int width = image1.cols;
  int height = image1.rows;

  GpuMat frame0GPU, frame1GPU, uGPU, vGPU;
  Mat image1_resize, image2_resize;
  Mat frame0, frame1, frame0_32, frame1_32, imgU, imgV;

  cv::gpu::BroxOpticalFlow dflow(alpha_,gamma_,scale_factor_,inner_iterations_,outer_iterations_,solver_iterations_);
  
  image1_resize = Mat(Size(cvRound(image1.cols/resizeFactor), cvRound(image1.rows/resizeFactor)), image1.type());
  image2_resize = Mat(Size(cvRound(image2.cols/resizeFactor), cvRound(image2.rows/resizeFactor)), image2.type());
  int width_resize = image1_resize.cols;
  int height_resize = image1_resize.rows;
  resize(image1, image1_resize, Size(width_resize,height_resize),0,0,INTER_LINEAR);
  resize(image2, image2_resize, Size(width_resize,height_resize),0,0,INTER_LINEAR);

  frame0 = Mat(Size(width_resize,height_resize), CV_8UC1);
  frame1 = Mat(Size(width_resize,height_resize), CV_8UC1);
  frame0_32 = Mat(Size(width_resize,height_resize), CV_32FC1);
  frame1_32 = Mat(Size(width_resize,height_resize), CV_32FC1);

  cvtColor(image1_resize, frame0, CV_BGR2GRAY);
  cvtColor(image2_resize, frame1 ,CV_BGR2GRAY);
  frame0.convertTo(frame0_32, CV_32FC1, 1.0/255.0,0);
  frame1.convertTo(frame1_32, CV_32FC1, 1.0/255.0,0);

  // Upload images to the GPU
  frame1GPU.upload(frame1_32);
  frame0GPU.upload(frame0_32);
  
  // Do the dense optical flow
  dflow(frame0GPU,frame1GPU,uGPU,vGPU);

  uGPU.download(imgU);
  vGPU.download(imgV);

  Mat imgU_originalSize;
  Mat imgV_originalSize;
  resize(imgU, imgU_originalSize, Size(width,height),0,0,INTER_NEAREST);
  resize(imgV, imgV_originalSize, Size(width,height),0,0,INTER_NEAREST);
  double ratioU = ((double)(width)) / ((double)(width_resize));
  double ratioV = ((double)(height)) / ((double)(height_resize));
  imgU_originalSize *= ratioU;
  imgV_originalSize *= ratioV;

  pair<Mat,Mat> output;
  output.first = imgU_originalSize;
  output.second = imgV_originalSize;

  return output;
}


/************************************************************************/
Mat affineEstimate(const Mat & P, const Mat & V){
  Mat A(3, 2, CV_32F);

  int nP = P.rows;

  Mat M(nP*2, 6, CV_32F);
  for(int i = 0; i < nP; i++){
    M.at<float>(i,0) = P.at<float>(i,0);
    M.at<float>(i,1) = P.at<float>(i,1);
    M.at<float>(i,2) = 1;
    M.at<float>(i,3) = 0;
    M.at<float>(i,4) = 0;
    M.at<float>(i,5) = 0;
    M.at<float>(nP+i,0) = 0;
    M.at<float>(nP+i,1) = 0;
    M.at<float>(nP+i,2) = 0;
    M.at<float>(nP+i,3) = P.at<float>(i,0);
    M.at<float>(nP+i,4) = P.at<float>(i,1);
    M.at<float>(nP+i,5) = 1;
  }

  Mat y(nP*2, 1, CV_32F);
  for(int i = 0; i < nP; i++){
    y.at<float>(i,0) = V.at<float>(i,0);
    y.at<float>(nP+i,0) = V.at<float>(i,1);
  }

  Mat B = Mat::eye(6, 6, CV_32F);
  B.at<float>(2,2) = 0;
  B.at<float>(5,5) = 0;

  Mat temp = M.t() * M + B;
  Mat tempInv = temp.inv();
  Mat x = tempInv * M.t() * y;

  A.at<float>(0,0) = x.at<float>(0,0);
  A.at<float>(1,0) = x.at<float>(1,0);
  A.at<float>(2,0) = x.at<float>(2,0);
  A.at<float>(0,1) = x.at<float>(3,0);
  A.at<float>(1,1) = x.at<float>(4,0);
  A.at<float>(2,1) = x.at<float>(5,0);
  
  return A;
}

/************************************************************************/
Mat affineMotionRansac(const Mat & P, const Mat & V, double thresh, int numIter, vector<int>& returnInliers){
  Mat A = Mat::zeros(3, 2, CV_32F);
  int K = 3;
  returnInliers.clear();
  
  int nP = P.rows;
  int nV = V.rows;
  if(nP != nV){
    cout << "affineMotionRansac: Matrix P and V have different number of elements." << endl;
    exit(1);
  }
  if(P.cols != 2 || V.cols != 2){
    cout << "affineMotionRansac: P and V both should have 2 columsn, X and Y." << endl;
    exit(1);
  }
  if(nP <= K){
    cout << "affineMotionRansac: Not enough elements in P." << endl;
  }
  
  srand(time(NULL));

  Mat Amax = Mat::zeros(3, 2, CV_32F);
  float cmax = 0;

  for(int iter = 1; iter <= numIter; iter++){
    vector<int> ind = randsample(nP, K);
    Mat Pind = indMatrixRows<float>(P, ind);
    Mat Vind = indMatrixRows<float>(V, ind);
    A = affineEstimate(Pind, Vind);
    
    Mat vy = P.colRange(0,1)*A.at<float>(0,0) + P.colRange(1,2)*A.at<float>(1,0) + A.at<float>(2,0);
    Mat vx = P.colRange(0,1)*A.at<float>(0,1) + P.colRange(1,2)*A.at<float>(1,1) + A.at<float>(2,1);

    Mat temp1 = vy - V.colRange(0,1);
    Mat temp2 = vx - V.colRange(1,2);
    temp1 = temp1.mul(temp1);
    temp2 = temp2.mul(temp2);
    Mat d = temp1 + temp2;
    int c = 0;
    for(int i = 0; i < d.rows; i++)
      if(d.at<float>(i,0) < thresh*thresh)
	c++;
    if(c > cmax){
      cmax = c;
      A.copyTo(Amax);
      returnInliers.clear();
      for(int i = 0; i < d.rows; i++)
	if(d.at<float>(i,0) < thresh*thresh)
	  returnInliers.push_back(i);
    }
  }

  Mat Pinliers = indMatrixRows<float>(P, returnInliers);
  Mat Vinliers = indMatrixRows<float>(V, returnInliers);

  A = affineEstimate(Pinliers, Vinliers);

  return A;
}

/*************************************************************************/
void getMotionChannels_spectralClustering(const Mat& VX, const Mat& VY, const int st, const int num, Mat& returnLABELS, SparseMat& returnAdj){
  int ysize = VX.rows;
  int xsize = VY.cols;
  Mat VYsub = subMatrix<float>(VY, st);
  Mat VXsub = subMatrix<float>(VX, st);

  Mat VXX, VXY, VYX, VYY;
  gradient(VYsub, VYY, VYX);
  gradient(VXsub, VXY, VXX);

  Mat temp = VXX.mul(VXX) + VYX.mul(VYX);
  Mat VH = powMatrix(temp, 0.5);
  temp = VXY.mul(VXY) + VYY.mul(VYY);
  Mat VV = powMatrix(temp, 0.5);

  int r = VXsub.rows;
  int c = VXsub.cols;
  SparseMat adj_r, adj_l, adj_u, adj_d;

  getImage4connectedAdjacencyMatrix(r, c, adj_r, adj_l, adj_u, adj_d);
  
  float coef = 10;
  float myEps = 0.00001;
  
  Mat VHr = VH.colRange(0, c-1) * coef;
  Mat VHl = VH.colRange(1, c) * coef;
  Mat VVu = VV.rowRange(1, r) * coef;
  Mat VVd = VV.rowRange(0, r-1) * coef;

  SparseMat::iterator adj_r_it = adj_r.begin();
  SparseMat::iterator adj_l_it = adj_l.begin();
  SparseMat::iterator adj_u_it = adj_u.begin();
  SparseMat::iterator adj_d_it = adj_d.begin();

  int dims = adj_r.dims();
  int idx[dims];
  const int * sizes = adj_r.size();
  int type = adj_r.type();
  SparseMat adj(dims, sizes, type);

  for(int i = 0; i < VHr.rows; i++){
    for(int j = 0; j < VHr.cols; j++){
      adj.ref<float>(adj_r_it.node()->idx) = myEps + exp(-VHr.at<float>(i,j));
      adj_r_it++;
    }
  }

  for(int i = 0; i < VHl.rows; i++){
    for(int j = 0; j < VHl.cols; j++){
      adj.ref<float>(adj_l_it.node()->idx) = myEps + exp(-VHl.at<float>(i,j));
      adj_l_it++;
    }
  }

  for(int i = 0; i < VVu.rows; i++){
    for(int j = 0; j < VVu.cols; j++){
      adj.ref<float>(adj_u_it.node()->idx) = myEps + exp(-VVu.at<float>(i,j));
      adj_u_it++;
    }
  }
  
  for(int i = 0; i < VVd.rows; i++){
    for(int j = 0; j < VVd.cols; j++){
      adj.ref<float>(adj_d_it.node()->idx) = myEps + exp(-VVd.at<float>(i,j));
      adj_d_it++;
    }
  }
  
  vector<int> labels = normalizedSpectralClustering(adj, num);
  Mat tempLABELS = Mat::zeros(r, c, CV_32S);
  int count = 0;
  for(int i = 0; i < r; i++)
    for(int j = 0; j < c; j++)
      tempLABELS.at<int>(i,j) = labels[count++];

  returnLABELS = Mat::zeros(ysize, xsize, CV_32S);
  resize(tempLABELS, returnLABELS, Size(xsize,ysize), 0, 0, INTER_NEAREST);

  return;
}

/*************************************************************************/
Mat getMotionChannels_clustering(const Mat& VX, const Mat& VY, const int st, const int num){
  int ysize = VX.rows;
  int xsize = VY.cols;
  Mat VYsub = subMatrix<float>(VY, st);
  Mat VXsub = subMatrix<float>(VX, st);

  float myepsilon = 0.000001;

  int h = VYsub.rows;
  int w = VYsub.cols;

  int N = h*w;

  Mat VY_ = VYsub.reshape(0, N);
  //cout << VY_.rows << " " << VY_.cols << endl;
  Mat VX_ = VXsub.reshape(0, N);
  VY_ = VY_.t();
  VX_ = VX_.t();
  VY_.push_back(VX_);
  VY_ = VY_.t();
  //cout << VY_.rows << " " << VY_.cols << endl;

  // clustering
  Mat idx(N, 1, CV_32S);
  for(int i = 0; i < N; i++)
    idx.at<int>(i,0) = -1;
  Mat centers(N, num, CV_32F);
  TermCriteria t(1, 100, myepsilon);
  kmeans(VY_, num, idx, t, 5, KMEANS_PP_CENTERS, centers);

  //cout << idx.rows << " " << idx.cols << endl;

  Mat IDX = idx.reshape(0, h);
  IDX += 1;

  //cout << IDX.rows << " " << IDX.cols << endl;

  Mat LABELS = Mat::zeros(ysize, xsize, CV_32S);
  resize(IDX, LABELS, Size(xsize,ysize), 0, 0, INTER_NEAREST);
  
  return LABELS;
}


/*************************************************************************/
Mat xiaofeng_getInliersOpticalFlowAffine(const Mat& im1, const Mat& im2, const int st){
  double resizeFactor = 2;
  pair<Mat,Mat> flow = getDenseFlow_eccv04(im1, im2, resizeFactor);
  return xiaofeng_getInliersOpticalFlowAffine(flow, st);
}

/*************************************************************************/
Mat xiaofeng_getInliersOpticalFlowAffine(const pair<Mat,Mat>& flow, const int st){
  Mat VX = flow.first;
  Mat VY = flow.second;
  int ysize = VY.rows;
  int xsize = VY.cols;

  // visualize optical flow
  /*Mat flow_rgb;
  flow_rgb = Mat(ysize, xsize, CV_8UC3);
  
  drawColorField(VX, VY, flow_rgb);
  
  imshow("Dense Flow",flow_rgb);*/
  
  // Computing motion channels
  int num = 3;
  Mat LABELS = getMotionChannels_clustering(VX, VY, st, num);

  // Finding the most frequent label (the biggest flow region)
  int freq[num];
  for(int i = 0; i < num; i++)
    freq[i] = 0;
  for(int i = 0 ; i < ysize; i++)
    for(int j = 0; j < xsize; j++)
      freq[LABELS.at<int>(i,j)-1]++;
  int I = -1;
  int maxF = -1;
  for(int i = 0; i < num; i++){
    if(freq[i] > maxF){
      I = i;
      maxF = freq[i];
    }
  }
  I = I + 1;

  // Average of optical flow in biggest region
  int count = 0;
  float meanVX = 0;
  float meanVY = 0;
  float meanV = 0;
  for(int i = 0 ; i < ysize; i++){
    for(int j = 0; j < xsize; j++){
      if(LABELS.at<int>(i,j) == I){
	count++;
	meanVY += VY.at<float>(i,j);
	meanVX += VX.at<float>(i,j);
	meanV += sqrt(pow(VY.at<float>(i,j), 2.0) + pow(VX.at<float>(i,j), 2.0));
      }
    }
  }
  meanVY /= (float)(count);
  meanVX /= (float)(count);
  meanV /= (float)(count);

  // Dividing VX and VY by meanV, and making them ready for RANSAC
  VY /= meanV;
  VX /= meanV;
  
  Mat VYsub = subMatrix<float>(VY, st);
  Mat VXsub = subMatrix<float>(VX, st);
  int h = VYsub.rows;
  int w = VYsub.cols;
  
  Mat VY_ = VYsub.reshape(0, h*w);
  Mat VX_ = VXsub.reshape(0, h*w);
  VY_ = VY_.t();
  VX_ = VX_.t();
  VY_.push_back(VX_);
  Mat V = VY_.t();
  
  Mat Xgrid(ysize, xsize, CV_32F);
  Mat Ygrid(ysize, xsize, CV_32F);
  for(int i = 0; i < ysize; i++){
    for(int j = 0; j < xsize; j++){
      Ygrid.at<float>(i,j) = (float)(i);
      Xgrid.at<float>(i,j) = (float)(j);
    }
  }
  Mat YgridSub = subMatrix<float>(Ygrid, st);
  Mat XgridSub = subMatrix<float>(Xgrid, st);
  Mat Ygrid_ = YgridSub.reshape(0, h*w);
  Mat Xgrid_ = XgridSub.reshape(0, h*w);
  Ygrid_ = Ygrid_.t();
  Xgrid_ = Xgrid_.t();
  Ygrid_.push_back(Xgrid_);
  Mat P = Ygrid_.t();

  // RANSAC to fit affine transformation

  vector<int> inliers;
  Mat A = affineMotionRansac(P, V, 0.2, 200, inliers); // magic numbers

  Mat mask(h, w, CV_32S);
  for(int i = 0; i < h; i++)
    for(int j = 0; j < w; j++)
      mask.at<int>(i,j) = 0;
  int y,x;
  for(vector<int>::iterator it = inliers.begin(); it != inliers.end(); it++){
    y = (int)(P.at<float>(*it, 0)/st);
    x = (int)(P.at<float>(*it, 1)/st);
    mask.at<int>(y,x) = 1;
  }

  Mat Mask(ysize, xsize, CV_32S);
  resize(mask, Mask, Size(xsize,ysize), 0, 0, INTER_NEAREST);

  return Mask;
}


/*************************************************************************/
Mat getBackgroundMaskPb(const vector<Mat>& images, const int refIndex){
  int N = images.size();
  if(refIndex >= N){
    cout << "getBackgroundMaskPb: refIndex is bigger than N" << endl;
    exit(1);
  }

  Mat refIM;
  images[refIndex].copyTo(refIM);
  int ysize = refIM.rows;
  int xsize = refIM.cols;

  Mat pbMask(ysize, xsize, CV_32F);
  for(int i = 0; i < ysize; i++)
    for(int j = 0; j < xsize; j++)
      pbMask.at<float>(i,j) = 0;

  int st = (7*ysize)/480; // magic number!

  for(int i = 0; i < N; i++){
    if(i == refIndex)
      continue;
    Mat mask = xiaofeng_getInliersOpticalFlowAffine(refIM, images[i], st);
    for(int i = 0; i < ysize; i++)
      for(int j = 0; j < xsize; j++)
	pbMask.at<float>(i,j) += (float)(mask.at<int>(i,j));
  }

  pbMask /= (N - 1);

  return pbMask;
}

/*************************************************************************/
Mat getBackgroundMaskPb(const vector<pair<Mat,Mat> >& flows){
  pair<Mat,Mat> flow0 = flows[0];
  Mat VX0 = flow0.first;
  int ysize = VX0.rows;
  int xsize = VX0.cols;

  Mat pbMask(ysize, xsize, CV_32F);
  for(int i = 0; i < ysize; i++)
    for(int j = 0; j < xsize; j++)
      pbMask.at<float>(i,j) = 0;

  int st = (7*ysize)/480; // magic number!

  int N = flows.size();

  for(int i = 0; i < N; i++){
    Mat mask = xiaofeng_getInliersOpticalFlowAffine(flows[i], st);
    for(int i = 0; i < ysize; i++)
      for(int j = 0; j < xsize; j++)
	pbMask.at<float>(i,j) += (float)(mask.at<int>(i,j));
  }

  pbMask /= N;

  return pbMask;
			
}

/*************************************************************************/
SparseMat getTransitionFlow_unnormalizedM(const Mat& VX, const Mat& VY, const Mat& segmentation1, const Mat& segmentation2){
  int y,x;
  int m = matrixMaxValue<int>(segmentation1, y, x); // number of regions 1
  int n = matrixMaxValue<int>(segmentation2, y, x); // number of regions 2

  Mat answer(n, m, CV_32F);
  for(int i = 0; i < n; i++)
    for(int j = 0; j < m; j++)
      answer.at<float>(i,j) = 0;

  int ysize = segmentation1.rows;
  int xsize = segmentation1.cols;

  int ii, jj, s1, s2;
  for(int i = 0; i < ysize; i++){
    for(int j = 0; j < xsize; j++){
      ii = i + (int)(VY.at<float>(i,j));
      jj = j + (int)(VX.at<float>(i,j));
      if(ii < 0 || jj < 0 || ii >= ysize || jj >= xsize)
	 continue;
      s2 = segmentation2.at<int>(ii,jj);
      s1 = segmentation1.at<int>(i,j);
      answer.at<float>(s2-1, s1-1)++;
    }
  } 

  SparseMat output(answer);
  return output;
}

/*************************************************************************/
/*************************************************************************/





