#include "PairwiseStuff.h"
#include "MotionStuff.h"

using namespace std;
using namespace cv;

/***************************************************************************/
Pairwise::Pairwise(const int frame1, const int frame2){
  f1_ = frame1;
  f2_ = frame2;

  computed_flow_ = false;
  computed_flow_unnormalized_ = false;
  computed_flow_normalized_ = false;
}

/***************************************************************************/
Mat Pairwise::getImage1(map<int, Prior>& priors) const{
  return priors[f1_].getImage();
}

/***************************************************************************/
Mat Pairwise::getImage2(map<int, Prior>& priors) const{
  return priors[f2_].getImage();
}

/***************************************************************************/
Mat Pairwise::getSegmentation1(map<int, Prior>& priors) const{
  return priors[f1_].getSegmentation();
}

/***************************************************************************/
Mat Pairwise::getSegmentation2(map<int, Prior>& priors) const{
  return priors[f2_].getSegmentation();
}

/***************************************************************************/
void Pairwise::setFlow(map<int, Prior>& priors, double resizeFactor=2){
  cout << "Computing dense flow " << f1_ << " " << f2_ << "...";
  cout.flush();

  flow_ = getDenseFlow_eccv04(priors[f1_].getImage(), priors[f2_].getImage(), resizeFactor);

  cout << "Done." << endl;

  computed_flow_ = true;
  computed_flow_unnormalized_ = false;
  computed_flow_normalized_ = false;

  return;
}

/***************************************************************************/
Mat Pairwise::getVX(map<int, Prior>& priors){
  if(computed_flow_)
    return flow_.first;

  setFlow(priors);
  
  return flow_.first;
}

/***************************************************************************/
Mat Pairwise::getVY(map<int, Prior>& priors){
  if(computed_flow_)
    return flow_.second;

  setFlow(priors);
  
  return flow_.second;
}

/***************************************************************************/
pair<Mat, Mat> Pairwise::getFlow(map<int, Prior>& priors){
  if(computed_flow_)
    return flow_;

  setFlow(priors);

  return flow_;
}

/***************************************************************************/
SparseMat Pairwise::getTransitionFlow_unnormalized(map<int, Prior>& priors){
  if(computed_flow_unnormalized_)
    return transitionFlow_unnormalized_;

  Mat segmentation1 = getSegmentation1(priors);
  Mat segmentation2 = getSegmentation2(priors);

  Mat VX = getVX(priors);
  Mat VY = getVY(priors);

  //cout << "Transition Unnormalized " << f1_ << " " << f2_ << "...";
  //cout.flush();

  transitionFlow_unnormalized_ = ::getTransitionFlow_unnormalizedM(VX, VY, segmentation1, segmentation2);
  
  //cout << "Done." << endl;

  computed_flow_unnormalized_ = true;
  return transitionFlow_unnormalized_;
}

/***************************************************************************/
SparseMat Pairwise::getTransitionFlow_normalized(map<int, Prior>& priors){
  if(computed_flow_normalized_)
    return transitionFlow_normalized_;

  float myepsilon = 0.0000001;

  SparseMat flowUnnorm = getTransitionFlow_unnormalized(priors);

  Mat S = Mat::ones(flowUnnorm.size()[0], 1, CV_32F) * myepsilon;
  for(SparseMat::iterator it = flowUnnorm.begin(); it != flowUnnorm.end(); it++)
    S.at<float>(it.node()->idx[0]) += it.value<float>();

  flowUnnorm.copyTo(transitionFlow_normalized_);
  for(SparseMat::iterator it = transitionFlow_normalized_.begin(); it != transitionFlow_normalized_.end(); it++)
    transitionFlow_normalized_.ref<float>(it.node()->idx) /= S.at<float>(it.node()->idx[0], 0);

  computed_flow_normalized_ = true;
  return transitionFlow_normalized_;  
}

/***************************************************************************/
Mat Pairwise::transferFrom1to2(const Mat& prior1, map<int, Prior>& priors){

  SparseMat flowUnnorm = getTransitionFlow_unnormalized(priors);

  float myepsilon = 0.00001;

  Mat S = Mat::zeros(flowUnnorm.size()[0], 1, CV_32F);
  for(SparseMat::iterator it = flowUnnorm.begin(); it != flowUnnorm.end(); it++)
    S.at<float>(it.node()->idx[0]) += it.value<float>();

  float avgPrior1 = 0;
  for(int i = 0; i < prior1.rows; i++){
    avgPrior1 += prior1.at<float>(i,0);
  }
  avgPrior1 /= (float)(prior1.rows);

  //cout << "Matrix multiplication " << f1_ << " " << f2_ << "...";
  //cout.flush();

  Mat transition;
  flowUnnorm.copyTo(transition);
  //cout << transition.rows << " " << transition.cols << endl;
  //cout << prior1.rows << " " << prior1.cols << endl;
  Mat prior2 = transition * prior1;
  
  //cout << "Done." << endl;

  float s;
  for(int i = 0; i < S.rows; i++){
    s = S.at<float>(i,0);
    if(abs(s) < myepsilon)
      prior2.at<float>(i,0) = avgPrior1;
    else
      prior2.at<float>(i,0) /= s;
  }
  
  return prior2;
}


/***************************************************************************/
SparseMat Pairwise::computeTemporal(map<int, Prior>& priors){
  return getTransitionFlow_unnormalized(priors);
}
/***************************************************************************/
/***************************************************************************/
/***************************************************************************/

