#include "PriorStuff.h"
#include "FeatureStuff.h"
#include "MotionStuff.h"
#include "SegmentationStuff.h"
#include "SparseStuff.h"

#include <vector>

using namespace std;
using namespace cv;

/*******************************************************************************/
Prior::Prior(){
  computed_size_ = false;
  computed_centers_ = false;
  computed_location_ = false;
  computed_motionChannel_ = false;
  computed_hogmap_ = false;
  computed_colormap_ = false;
  return;
}

/*******************************************************************************/
Prior::Prior(const Prior& p){
  image_ = p.getImage();
  segmentation_ = p.getSegmentation();
  adjacencyLengths_ = p.getAdjacencyLengths();
  adjacencyWeights_ = p.getAdjacencyWeights();
  
  f_ = p.getFrameNumber();
  m_ = p.getM();
  computed_size_ = p.isSizeComputed();
  computed_centers_ = p.isCentersComputed();
  computed_location_ = p.isLocationComputed();
  computed_motionChannel_ = p.isMotionChannelComputed();
  computed_hogmap_ = p.isHogMapComputed();
  computed_colormap_ = p.isColorMapComputed();
  if(computed_centers_)
    centers_ = p.copyCenters();
  if(computed_size_)
    prior_size_ = p.copyPrior_size();
  if(computed_location_)
    prior_location_ = p.copyPrior_location();
  if(computed_motionChannel_)
    prior_motionChannel_ = p.copyPrior_motionChannel();
  if(computed_colormap_)
    colorMap_ = p.copyColorMap();
  if(computed_hogmap_)
    hogMap_ = p.copyHogMap();

  return;
}

/*******************************************************************************/
Prior::Prior(const Mat& image, const int frameNumber){
  image.copyTo(image_);

  // setting booleans
  computed_centers_ = false;
  computed_size_ = false;
  computed_location_ = false;
  computed_motionChannel_ = false;
  computed_colormap_ = false;
  computed_hogmap_ = false;

  // segmentation
  float sigma = 0.4;
  float k = 100;
  int min_size = 100;
  int num_ccs;
  Mat segmentation = felzenszwalb_segment(image, sigma, k, min_size, &num_ccs);
  sortSegmentNumbers(segmentation);
  segmentation.copyTo(segmentation_);

  adjacencyLengths_ = getWeightedAdjacencyMatrix(segmentation);
  int binSize = 16;
  vector<vector<float> > featureVectors = getFeatureVectors_colorHist(image, binSize, segmentation);
  float var1 = 1;
  float var2 = 1;
  adjacencyWeights_ = getSpatial_colorHist(featureVectors, adjacencyLengths_, var1, var2);

  int y,x;
  m_ = matrixMaxValue<int>(segmentation_, y, x); // number of regions
  f_ = frameNumber; // frame number
  
  return;
}

/*******************************************************************************/
Prior::Prior(const Mat& image, const Mat& ucm, const int k, const int frameNumber){
  image.copyTo(image_);

  // setting booleans
  computed_centers_ = false;
  computed_size_ = false;
  computed_location_ = false;
  computed_motionChannel_ = false;
  computed_colormap_ = false;
  computed_hogmap_ = false;

  // segmentation
  segmentation_ = berkeley_segment(ucm, k);
  sortSegmentNumbers(segmentation_);

  SparseMat tempAdjWeights;
  berkeley_weightedAdjacencyMatrix(ucm, segmentation_, adjacencyLengths_, tempAdjWeights);
  float var1 = 4.0;
  float var2 = 0.5;
  adjacencyWeights_ = applyVarAdjacency(tempAdjWeights, var1, var2);

  int y,x;
  m_ = matrixMaxValue<int>(segmentation_, y, x); // number of regions
  f_ = frameNumber; // frame number
  
  return;
}

/*******************************************************************************/
void Prior::setFrameNumber(const int frameNumber){
  f_ = frameNumber;
  return;
}

/*******************************************************************************/
int Prior::getFrameNumber() const{
  return f_;
}

/*******************************************************************************/
void Prior::setSegmentation(const Mat& segmentation){
  segmentation.copyTo(segmentation_);
  sortSegmentNumbers(segmentation_);

  computed_centers_ = false;
  computed_size_ = false;
  computed_location_ = false;
  computed_motionChannel_ = false;
  computed_hogmap_ = false;
  computed_colormap_ = false;

  return;
}

/*******************************************************************************/
void Prior::setSegmentation_felz(const Mat& segmentation){
  segmentation.copyTo(segmentation_);
  sortSegmentNumbers(segmentation_);

  adjacencyLengths_ = getWeightedAdjacencyMatrix(segmentation);
  int binSize = 16;
  vector<vector<float> > featureVectors = getFeatureVectors_colorHist(image_, binSize, segmentation);
  float var1 = 1;
  float var2 = 1;
  adjacencyWeights_ = getSpatial_colorHist(featureVectors, adjacencyLengths_, var1, var2);

  int y,x;
  m_ = matrixMaxValue<int>(segmentation_, y, x); // number of regions

  computed_centers_ = false;
  computed_size_ = false;
  computed_location_ = false;
  computed_motionChannel_ = false;
  computed_hogmap_ = false;
  computed_colormap_ = false;
  
  return;
}

/*******************************************************************************/
void Prior::setSegmentation_berkeley(const Mat& ucm, const int k){
  segmentation_ = berkeley_segment(ucm, k);
  sortSegmentNumbers(segmentation_);

  SparseMat tempAdjWeights;
  berkeley_weightedAdjacencyMatrix(ucm, segmentation_, adjacencyLengths_, tempAdjWeights);
  float var1 = 4.0;
  float var2 = 0.5;
  adjacencyWeights_ = applyVarAdjacency(tempAdjWeights, var1, var2);
  
  int y,x;
  m_ = matrixMaxValue<int>(segmentation_, y, x); // number of regions

  computed_centers_ = false;
  computed_size_ = false;
  computed_location_ = false;
  computed_motionChannel_ = false;
  computed_colormap_ = false;
  computed_hogmap_ = false;
  
  return;
}

/*******************************************************************************/
Mat Prior::getImage() const{
  return image_;
}

/*******************************************************************************/
Mat Prior::getSegmentation() const{
  return segmentation_;
}

/*******************************************************************************/
SparseMat Prior::getAdjacencyWeights() const{
  return adjacencyWeights_;
}

/*******************************************************************************/
SparseMat Prior::getAdjacencyLengths() const{
  return adjacencyLengths_;
}

/*******************************************************************************/
SparseMat Prior::getW() const{
  return getSpatialWeighted(adjacencyWeights_, adjacencyLengths_);
}

/*******************************************************************************/
int Prior::getNumSegments() const{
  return m_;
}

/*******************************************************************************/
Mat Prior::getCenters(){
  if(computed_centers_)
    return centers_;

  vector< pair<float,float> > centers = getSegmentCenters(segmentation_);
  int N = (int)(centers.size());
  centers_ = Mat::zeros(N, 2, CV_32F);
  int count = 0;
  for(vector< pair<float,float> >::iterator it = centers.begin(); it != centers.end(); it++){
    centers_.at<float>(count,0) = it->first;
    centers_.at<float>(count,1) = it->second;
    count++;
  }
  computed_centers_ = true;

  return centers_;
}

/*******************************************************************************/
Mat Prior::getPrior_size(){
  if(computed_size_)
    return prior_size_;

  prior_size_ = Mat::zeros(m_, 1, CV_32F);

  int h = segmentation_.rows;
  int w = segmentation_.cols;
  for(int i = 0; i < h; i++)
    for(int j = 0; j < w; j++)
      prior_size_.at<float>(segmentation_.at<int>(i,j)-1, 0)++;

  computed_size_ = true;
  return prior_size_;
}

/*******************************************************************************/
Mat Prior::getPrior_location(const Mat& locationPattern){
  if(computed_location_)
    return prior_location_;

  Mat centers = getCenters();
  prior_location_ = Mat::zeros(m_, 1, CV_32F);

  int y,x;
  for(int i = 0; i < m_; i++){
    y = (int)(centers.at<float>(i,0));
    x = (int)(centers.at<float>(i,1));
    prior_location_.at<float>(i,0) = locationPattern.at<float>(y,x);
  }

  computed_location_ = true;  
  return prior_location_;
}

/*******************************************************************************/
Mat Prior::getPrior_motionChannel(map<int, Prior>& priors, const vector<int>& adjFrames){
  if(computed_motionChannel_)
    return prior_motionChannel_;


  vector<Mat> temp;
  temp.push_back(image_);
  for(int i = 0; i < adjFrames.size(); i++){
    temp.push_back(priors[adjFrames[i]].getImage());
  }
  Mat maskPb = getBackgroundMaskPb(temp, 0);

  Mat centers = getCenters();
  int N = centers.rows;
  prior_motionChannel_ = Mat::zeros(N, 1, CV_32F);
  int y,x;
  for(int i = 0; i < N; i++){
    y = (int)(centers.at<float>(i,0));
    x = (int)(centers.at<float>(i,1));
    prior_motionChannel_.at<float>(i,0) = maskPb.at<float>(y,x);
  }

  computed_motionChannel_ = true;

  return prior_motionChannel_;
}

/*******************************************************************************/
Mat Prior::getPrior_motionChannel(const vector<pair<Mat,Mat> >& flows){
  if(computed_motionChannel_)
    return prior_motionChannel_;

  Mat maskPb = getBackgroundMaskPb(flows);
  
  Mat centers = getCenters();
  int N = centers.rows;
  prior_motionChannel_ = Mat::zeros(N, 1, CV_32F);
  int y,x;
  for(int i = 0; i < N; i++){
    y = (int)(centers.at<float>(i,0));
    x = (int)(centers.at<float>(i,1));
    prior_motionChannel_.at<float>(i,0) = maskPb.at<float>(y,x);
  }

  computed_motionChannel_ = true;

  return prior_motionChannel_;
}

/*******************************************************************************/
Mat Prior::maskToLabel(const Mat& mask){
  Mat output(m_, 1, CV_32F);
  Mat centers = getCenters();
  int y,x;
  for(int i = 0; i < m_; i++){
    y = (int)(centers.at<float>(i,0));
    x = (int)(centers.at<float>(i,1));
    if(mask.at<uchar>(y,x) > 0)
      output.at<float>(i,0) = 1;
    else
      output.at<float>(i,0) = -1;
  }
  return output;
}

/*******************************************************************************/
Mat Prior::labelsToMask(const Mat& labels){
  int h = segmentation_.rows;
  int w = segmentation_.cols;

  Mat Labels(h, w, CV_32S);
  int s;
  for(int i = 0; i < h; i++){
    for(int j = 0; j < w; j++){
      s = segmentation_.at<int>(i,j);
      Labels.at<int>(i,j) = labels.at<int>(s-1,0);
    }
  }

  return Labels;
}

/*******************************************************************************/
Mat Prior::getColorMap(const cv::Mat& clusterCenters){
  if(computed_colormap_)
    return colorMap_;

  int numBins = clusterCenters.rows;
  Mat fv_lab = getColorLab(image_);
  Mat bins_lab = getFeatureMap(fv_lab, clusterCenters);
  colorMap_ = featureMapLab_reshape(bins_lab, image_.rows, image_.cols);
  computed_colormap_ = true;

  return colorMap_;
}

/*******************************************************************************/
Mat Prior::getHogMap(const cv::Mat& clusterCenters){
  if(computed_hogmap_)
    return hogMap_;

  int numBins = clusterCenters.rows;
  Mat fv_hog = getHOG(image_);
  Mat bins_hog = getFeatureMap(fv_hog, clusterCenters);
  hogMap_ = featureMapHOG_reshape(bins_hog, image_.rows, image_.cols);
  computed_hogmap_ = true;

  return hogMap_;
}

/*******************************************************************************/
Mat Prior::getFeature_CLab(const Mat& clusterCenters){
  int numBins = clusterCenters.rows;
  Mat featureMap_lab = getColorMap(clusterCenters);
  Mat fv_regions = featureMap2regionFeatures(featureMap_lab, segmentation_, numBins);
  return fv_regions;
}

/*******************************************************************************/
Mat Prior::getFeature_HOG(const Mat& clusterCenters){
  int numBins = clusterCenters.rows;
  Mat featureMap_hog = getHogMap(clusterCenters);
  Mat fv_regions = featureMap2regionFeatures(featureMap_hog, segmentation_, numBins);
  return fv_regions;
}

/*******************************************************************************/
Mat Prior::getFeatureAroundPoint_CLab(const Mat& clusterCenters, const float cy, const float cx, const int radius){
  int numBins = clusterCenters.rows;
  Mat featureMap_lab = getColorMap(clusterCenters);
  Mat output = featureAroundPoint(featureMap_lab, cy, cx, radius, numBins);
  return output;
}

/*******************************************************************************/
Mat Prior::getFeatureAroundPoint_HOG(const Mat& clusterCenters, const float cy, const float cx, const int radius){
  int numBins = clusterCenters.rows;
  Mat featureMap_hog = getHogMap(clusterCenters);
  Mat output = featureAroundPoint(featureMap_hog, cy, cx, radius, numBins);
  return output;
}

/*******************************************************************************/
Mat Prior::getFeatureAroundPoint_colorHOG(const Mat& clusterCenters_lab, const Mat& clusterCenters_hog, const float cy, const float cx, const int radius){
  Mat F1 = getFeatureAroundPoint_CLab(clusterCenters_lab, cy, cx, radius);
  Mat F2 = getFeatureAroundPoint_HOG(clusterCenters_hog, cy, cx, radius);
  F1 = F1.t();
  F2 = F2.t();
  F1.push_back(F2);
  F1 = F1.t();
  return F1;
}

/*******************************************************************************/
Mat Prior::getFeatureInMask_CLab(const Mat& clusterCenters, const Mat& mask){
  int numBins = clusterCenters.rows;
  Mat featureMap_lab = getColorMap(clusterCenters);
  Mat output = featureInMask(featureMap_lab, mask, numBins);
  //cout << "Mask_CLab: " << output.rows << " " << output.cols << endl;
  return output;
}
/*******************************************************************************/
Mat Prior::getFeatureInMask_HOG(const Mat& clusterCenters, const Mat& mask){
  int numBins = clusterCenters.rows;
  Mat featureMap_hog = getHogMap(clusterCenters);
  Mat output = featureInMask(featureMap_hog, mask, numBins);
  //cout << "Mask_HOG: " << output.rows << " " << output.cols << endl;
  return output;
}
/*******************************************************************************/
Mat Prior::getFeatureInMask_colorHOG(const Mat& clusterCenters_lab, const Mat& clusterCenters_hog, const Mat& mask){
  Mat F1 = getFeatureInMask_CLab(clusterCenters_lab, mask);
  Mat F2 = getFeatureInMask_HOG(clusterCenters_hog, mask);
  F1 = F1.t();
  F2 = F2.t();
  F1.push_back(F2);
  F1 = F1.t();
  return F1;
}

/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
PriorH::PriorH(const cv::Mat& image, const cv::Mat& ucm, const int k, const int k2, const int k3, const int frameNumber): Prior(image, ucm, k, frameNumber){
  segmentation2_ = berkeley_segment(ucm, k2);
  sortSegmentNumbers(segmentation2_);
  segmentation3_ = berkeley_segment(ucm, k3);
  sortSegmentNumbers(segmentation3_);
  return;
}

/*******************************************************************************/
Mat PriorH::getHierarchicalFeature_CLab(const Mat& clusterCenters){
  int numBins = clusterCenters.rows;
  Mat fv_lab = getColorLab(image_);
  Mat bins = getFeatureMap(fv_lab, clusterCenters);
  Mat featureMap = featureMapLab_reshape(bins, image_.rows, image_.cols);
  Mat fv = featureMap2regionFeatures(featureMap, segmentation_, numBins);
  Mat fv2 = featureMap2regionFeatures(featureMap, segmentation2_, numBins);
  Mat fv3 = featureMap2regionFeatures(featureMap, segmentation3_, numBins);

  Mat centers = getCenters();

  Mat fvh2 = ::getHierarchicalFeature(fv2, centers, segmentation2_);
  Mat fvh3 = ::getHierarchicalFeature(fv3, centers, segmentation3_);

  Mat output = fv.t();
  Mat fvh2t = fvh2.t();
  output.push_back(fvh2t);
  Mat fvh3t = fvh3.t();
  output.push_back(fvh3t);
  output = output.t();
  return output;
}

/*******************************************************************************/
Mat PriorH::getHierarchicalFeature_HOG(const Mat& clusterCenters){
  int numBins = clusterCenters.rows;
  Mat fv_hog = getHOG(image_);
  Mat bins = getFeatureMap(fv_hog, clusterCenters);
  Mat featureMap = featureMapHOG_reshape(bins, image_.rows, image_.cols);
  Mat fv = featureMap2regionFeatures(featureMap, segmentation_, numBins);
  Mat fv2 = featureMap2regionFeatures(featureMap, segmentation2_, numBins);
  Mat fv3 = featureMap2regionFeatures(featureMap, segmentation3_, numBins);

  Mat centers = getCenters();

  Mat fvh2 = getHierarchicalFeature(fv2, centers, segmentation2_);
  Mat fvh3 = getHierarchicalFeature(fv3, centers, segmentation3_);

  Mat output = fv.t();
  Mat fvh2t = fvh2.t();
  output.push_back(fvh2t);
  Mat fvh3t = fvh3.t();
  output.push_back(fvh3t);
  output = output.t();
  return output;  
}

/*******************************************************************************/
Mat PriorH::getHierarchicalFeature_colorHOG(const Mat& clusterCenters_lab, const Mat& clusterCenters_hog){
  Mat fv_lab = getHierarchicalFeature_CLab(clusterCenters_lab);
  Mat fv_hog = getHierarchicalFeature_HOG(clusterCenters_hog);
  Mat F = fv_lab.t();
  Mat fv_hogt = fv_hog.t();
  F.push_back(fv_hogt);
  F = F.t();
  return F;
}
/*******************************************************************************/
/*******************************************************************************/
