#include "SequenceStuff.h"
#include "InputOutputFile.h"
#include "MyMATFileInterface.h"
#include "SparseStuff.h"

#include <stdio.h>

using namespace std;
using namespace cv;

/************************************************************************/
Sequence::Sequence(){
  priors_.clear();
  pairwises_.clear();
  startFrame_ = -1;
  endFrame_ = -1;
}

/************************************************************************/
Sequence::Sequence(Dataset& db, const int seqNumber, const int startFrame, const int endFrame, const int seg_type){
  Mat im;
  cout << "Loading frames " << startFrame << " to " << endFrame << endl;
  for(int f = startFrame; f <= endFrame; f++){
    cout << ".";
    cout.flush();
    im = db.getIM(seqNumber, f);
    
    if(seg_type == FELZ_SEG){
      Prior p(im, f);
      push_back(p);
    }

    if(seg_type == BERKELEY_SEG){
      Mat ucm = db.readUCM(seqNumber, f);
      int k = 20;
      Prior p(im, ucm, k, f);
      push_back(p);
    }
    
  }
  cout << endl;

  return;
}

/************************************************************************/
void Sequence::push_back(const Prior& p){
  if(priors_.size() >= 1){
    Prior p_ = p;
    p_.setFrameNumber(endFrame_+1);
    
    Pairwise pp1(endFrame_, endFrame_+1);
    pairwises_[pair<int,int>(endFrame_,endFrame_+1)] = pp1;
    Pairwise pp2(endFrame_+1, endFrame_);
    pairwises_[pair<int,int>(endFrame_+1,endFrame_)] = pp2;
    endFrame_++;
  }
  else{
    startFrame_ = p.getFrameNumber();
    endFrame_ = p.getFrameNumber();
  }

  priors_[endFrame_] = p;

  return;
}

/************************************************************************/
void Sequence::push_back(Dataset& db, const int seqNumber, const int seg_type){

  int frame = endFrame_ + 1;
  Mat im = db.getIM(seqNumber, frame);

  if(seg_type == FELZ_SEG){
    Prior p(im, frame);
    push_back(p);
  }

  if(seg_type == BERKELEY_SEG){
    Mat ucm = db.readUCM(seqNumber, frame);
    int k = 20;
    Prior p(im, ucm, k, frame);
    push_back(p);
  }
  return;
}

/************************************************************************/
void Sequence::push_front(const Prior& p){
  if(priors_.size() >= 1){
    Prior p_ = p;
    p_.setFrameNumber(startFrame_-1);
    
    Pairwise pp1(startFrame_, startFrame_-1);
    pairwises_[pair<int,int>(startFrame_,startFrame_-1)] = pp1;
    Pairwise pp2(startFrame_-1, startFrame_);
    pairwises_[pair<int,int>(startFrame_-1,startFrame_)] = pp2;
    startFrame_--;
  }
  else{
    startFrame_ = p.getFrameNumber();
    endFrame_ = p.getFrameNumber();
  }

  priors_[endFrame_] = p;

  return;
}

/************************************************************************/
Prior Sequence::getPrior(int frameNumber){
  return priors_[frameNumber];
}

/************************************************************************/
Pairwise Sequence::getPairwise(int frameNumber1, int frameNumber2){
  return pairwises_[pair<int,int>(frameNumber1,frameNumber2)];
}

/************************************************************************/
void Sequence::pop_back(){
  if(priors_.size() < 1)
    return;
  priors_.erase(endFrame_);
  pairwises_.erase(pair<int,int>(endFrame_,endFrame_-1));
  pairwises_.erase(pair<int,int>(endFrame_-1,endFrame_));
  endFrame_--;
  return;
}

/************************************************************************/
void Sequence::pop_front(){
  if(priors_.size() < 1)
    return;
  priors_.erase(startFrame_);
  pairwises_.erase(pair<int,int>(startFrame_,startFrame_+1));
  pairwises_.erase(pair<int,int>(startFrame_+1,startFrame_));
  startFrame_++;
  return;
}

/************************************************************************/
int Sequence::getLength() const{
  return endFrame_-startFrame_+1;
}

/************************************************************************/
Mat Sequence::transferPriorFromFrameToFrame(const Mat& prior1, const int frame1, const int frame2){
  Mat output;
  if(frame1 == frame2){
    prior1.copyTo(output);
    return output;
  }

  int temp = 1;
  if(frame2 < frame1)
    temp = -1;

  //cout << "Transfering from " << frame1 << " to " << frame2 << endl; 

  prior1.copyTo(output);
  for(int i = frame1; i != frame2; i+=temp)
    output = pairwises_[pair<int,int>(i,i+temp)].transferFrom1to2(output, priors_);
  
  return output;
}

/************************************************************************/
Mat Sequence::xiaofengFeatureVector(const int frame, const Mat& locationPattern){
  Prior p = priors_[frame];
  Mat output = p.getPrior_location(locationPattern);

  vector<int> margins;
  margins.push_back(0);
  margins.push_back(-1);
  margins.push_back(-3);
  margins.push_back(-5);
  margins.push_back(-8);
  
  output = output.t();
  for(int i = 0; i < margins.size(); i++){
    int f = frame + margins[i];
    vector<int> adjFrames;
    adjFrames.push_back(f-1);
    adjFrames.push_back(f+1);
    vector<pair<Mat,Mat> > flows;
    for(int j = 0; j < adjFrames.size(); j++)
      flows.push_back(pairwises_[pair<int,int>(f,adjFrames[j])].getFlow(priors_));
    //cout << "MotionPrior...";
    //cout.flush();
    Mat motionPrior = priors_[f].getPrior_motionChannel(flows);
    //cout << "Done." << endl;
    Mat tfMotionPrior = transferPriorFromFrameToFrame(motionPrior, f, frame);
    //char buffer[100];
    //buffer[0] = 0;
    //sprintf(buffer, "motionPrior %d", margins[i]);
    //Mat seg = p.getSegmentation();
    //showPrior(tfMotionPrior, seg, buffer);
    Mat temp = tfMotionPrior.t();
    output.push_back(temp);
    
  }
  output = output.t();

  return output;
}

/************************************************************************/
SparseMat Sequence::computeW(const int fromFrame, const int toFrame, const float ratio){
  if(fromFrame < startFrame_ || toFrame > endFrame_){
    cout << "computeW: fromFrame or toFrame go beyond the frames contained in this sequence" << endl;
    exit(1);
  }

  vector<int> ms;
  for(int f=fromFrame; f <= toFrame; f++)
    ms.push_back(getPrior(f).getM());

  int m = 0;
  for(int i = 0; i < ms.size(); i++)
    m += ms[i];

  int dims = 2;
  int sizes[2];
  sizes[0] = m;
  sizes[1] = m;
  SparseMat output(dims, sizes, CV_32F);

  float ratioRev = 1.0/ratio;

  // setting spatial
  cout << "Building Spatial Matrix...";
  cout.flush();
  int m_start = 0;
  int m_;
  for(int i = 0; i < ms.size(); i++){
    m_ = ms[i];
    Prior p = getPrior(fromFrame+i);
    SparseMat sp = p.getW();
    int idx[2];
    for(SparseMat::iterator it = sp.begin(); it != sp.end(); it++){
      idx[0] = it.node()->idx[0] + m_start;
      idx[1] = it.node()->idx[1] + m_start;
      output.ref<float>(idx) = it.value<float>() * ratioRev;
    }
    m_start = m_start + m_;
  }
  cout << "Done." << endl;
  

  // setting temporal
  cout << "Building Temporal Matrix...";
  cout.flush();
  m_start = 0;
  int m_end = 0;
  int m1_, m2_;
  for(int i = 0; i < ms.size()-1; i++){
    m1_ = ms[i];
    m2_ = ms[i+1];
    m_end = m_start + m1_;
    SparseMat tp12 = pairwises_[pair<int,int>(fromFrame+i+1,fromFrame+i)].computeTemporal(priors_);
    SparseMat tp21 = pairwises_[pair<int,int>(fromFrame+i,fromFrame+i+1)].computeTemporal(priors_);
    //cout << "m1 m2: " << m1_ << " " << m2_ << endl;
    Pairwise p12 = getPairwise(fromFrame+i,fromFrame+i+1);
    Pairwise p21 = getPairwise(fromFrame+i+1,fromFrame+i);
    int idx[2];
    for(SparseMat::iterator it = tp12.begin(); it != tp12.end(); it++){
      idx[0] = it.node()->idx[0] + m_start;
      idx[1] = it.node()->idx[1] + m_end;
      output.ref<float>(idx) = it.value<float>();
    }
    for(SparseMat::iterator it = tp21.begin(); it != tp21.end(); it++){
      idx[0] = it.node()->idx[0] + m_end;
      idx[1] = it.node()->idx[1] + m_start;
      output.ref<float>(idx) = it.value<float>();
    }
    m_start = m_end;
  }
  cout << "Done." << endl;
  
  //printMatrix<float>(output);

  return output;
}


/************************************************************************/
