#include "SparseStuff.h"

#include <boost/numeric/ublas/matrix_sparse.hpp>
#include <boost/numeric/ublas/operation.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <ietl/interface/ublas.h>
#include <ietl/vectorspace.h>
#include <ietl/lanczos.h>
#include <ietl/iteration.h>
#include <boost/random.hpp>
#include <boost/limits.hpp>
#include <cmath>
#include <limits>

#include <vector>

typedef boost::numeric::ublas::compressed_matrix<float, boost::numeric::ublas::row_major>  SparseMatrix;
typedef boost::numeric::ublas::vector<float> VectorBlas;

using namespace std;
using namespace cv;

/****************************************************************************/
pair<Mat, Mat> eigs(const SparseMat & symA, const int k, const int type){
  int r, c;
  r = symA.size()[0];
  c = symA.size()[1];
  if(r != c){
    cout << "eigs: The matrix is not symmetric" << endl;
    exit(1);
  }

  int N = r;

  int nnzc = symA.nzcount();

  // Converting SparseMat symA to ublas compressed matrix
  SparseMatrix mat(N, N, nnzc);
  for(SparseMat::const_iterator it = symA.begin(); it != symA.end(); it++){
    int i = it.node()->idx[0];
    int j = it.node()->idx[1];
    mat(i,j) = it.value<float>();
  }

  //cout << "A:" << endl << A << endl;

  typedef ietl::wrapper_vectorspace<VectorBlas> Vecspace;
  typedef boost::lagged_fibonacci607 Gen; 
  Vecspace vec(N);
  Gen mygen;
  ietl::lanczos<SparseMatrix,Vecspace> lanczos(mat,vec);

  // Creation of an iteration object:    
  int max_iter = min(10*N, 300);
  float rel_tol = 500*std::numeric_limits<float>::epsilon();
  float abs_tol = pow(std::numeric_limits<float>::epsilon(),2./3.);  

  //std::cout << "Calculation of " << k << " lowest converged eigenvalues" << endl;
  //std::cout << "-----------------------------------" << endl;
  
  vector<float> eigen;
  vector<float> err;
  vector<int> multiplicity;
  
  if(type == EIGS_SMALLEST){
    ietl::lanczos_iteration_nlowest<float> iter(max_iter, k, rel_tol, abs_tol);
    try {
      lanczos.calculate_eigenvalues(iter,mygen);
      eigen = lanczos.eigenvalues();
      err = lanczos.errors();
      multiplicity = lanczos.multiplicities();
    }
    catch (std::runtime_error& e) {
      std::cout << e.what() << std::endl;
    }
  }

  if(type == EIGS_LARGEST){
    ietl::lanczos_iteration_nhighest<float> iter(max_iter, k, rel_tol, abs_tol);
    try {
      lanczos.calculate_eigenvalues(iter,mygen);
      eigen = lanczos.eigenvalues();
      err = lanczos.errors();
      multiplicity = lanczos.multiplicities();
    }
    catch (std::runtime_error& e) {
      std::cout << e.what() << std::endl;
    }
  }

  Mat output_values(k, 1, CV_32F);
  Mat output_vectors(k, N, CV_32F);

  // Printing eigenvalues with error & multiplicities:  
  /*cout << "#        eigenvalue            error         multiplicity" << endl;
  cout.precision(10);
  for (int i = 0; i < eigen.size(); ++i)
  cout << i << "\t" << eigen[i] << "\t" << err[i] << "\t" << multiplicity[i] << endl;*/
  
  if(type == EIGS_SMALLEST)
    for(int i = 0; i < k; i++)
      output_values.at<float>(i,0) = eigen[i];

  if(type == EIGS_LARGEST)
    for(int i = N-k; i < N; i++)
      output_values.at<float>(i-N+k,0) = eigen[i];
  
  // call of eigenvectors function follows:   
  //cout << "\nEigen vectors computations for " << k << " lowest eigenvalues" << endl;  

  if(type == EIGS_SMALLEST){
    vector<float>::iterator start = eigen.begin();
    vector<float>::iterator end = eigen.begin() + k;
    vector<VectorBlas> eigenvectors; // for storing the eigen vectors. 
    ietl::Info<float> info; // (m1, m2, ma, eigenvalue, residual, status).
    
    try {
      lanczos.eigenvectors(start, end, back_inserter(eigenvectors), info, mygen); 
    }
    catch (runtime_error& e) {
      cout << e.what() << endl;
    }  
    
    for(int i = 0; i < k; i++){
      VectorBlas temp = eigenvectors[i];
      int j = 0;
      for(VectorBlas::iterator it = temp.begin(); it != temp.end(); it++)
	output_vectors.at<float>(i,j++) = *it;
    }
  }


  if(type == EIGS_LARGEST){
    vector<float>::iterator start = eigen.begin()+(N-k-1);
    vector<float>::iterator end = eigen.begin()+(N-1);
    vector<VectorBlas> eigenvectors; // for storing the eigen vectors. 
    ietl::Info<float> info; // (m1, m2, ma, eigenvalue, residual, status).
    
    try {
      lanczos.eigenvectors(start, end, back_inserter(eigenvectors), info, mygen); 
    }
    catch (runtime_error& e) {
      cout << e.what() << endl;
    }  
    
    for(int i = 0; i < k; i++){
      VectorBlas temp = eigenvectors[i];
      int j = 0;
      for(VectorBlas::iterator it = temp.begin(); it != temp.end(); it++)
	output_vectors.at<float>(i,j++) = *it;
    }
  }

  
  /*cout << "Printing eigen Vectors" << endl; 
  for(vector<VectorBlas>::iterator it = eigenvectors.begin(); it != eigenvectors.end(); it++){
    copy((it)->begin(),(it)->end(),ostream_iterator<float>(cout,"\n"));
    cout << endl;
    }*/

  pair<Mat, Mat> output;
  output.first = output_values;
  output.second = output_vectors;

  //cout << "Values: " << endl << output.first << endl;
  //cout << "Vectors: " << endl << output.second << endl;
  return output;
}

/****************************************************************************/
SparseMat getNormalizedLaplacianMatrix(const SparseMat& symA){
  if(symA.size()[0] != symA.size()[1]){
    cout << "getLaplacianMatrix: symA is not a square matrix" << endl;
    exit(1);
  }
  int N = symA.size()[0];

  cout << "Nonzero elements in symA: " << symA.nzcount() << endl;
  cout << "Copying symA to A" << endl;

  SparseMatrix A(N, N);
  for(SparseMat::const_iterator it = symA.begin(); it != symA.end(); it++){
    int i = it.node()->idx[0];
    int j = it.node()->idx[1];
    A(i,j) = it.value<float>();
  }

  SparseMat symD = matrixToD<float>(symA);

  cout << "Computing Dnim_" << endl;

  cv::SparseMat Dnim_(symD.dims(), symD.size(), symD.type());
  for(cv::SparseMat::const_iterator it = symD.begin(); it != symD.end(); it++)
    Dnim_.ref<float>(it.node()->idx) = 1.0 / sqrt(it.value<float>());

  cout << "Nonzero elements in Dnim_: " << Dnim_.nzcount() << endl;

  cout << "Copying Dnim_ to Dnim" << endl;

  SparseMatrix Dnim(N,N);
  for(SparseMat::const_iterator it = Dnim_.begin(); it != Dnim_.end(); it++){
    int i = it.node()->idx[0];
    int j = it.node()->idx[1];
    Dnim(i,j) = it.value<float>();
  }

  cout << "Creaintg I" << endl;
  
  SparseMatrix I(N, N, N);
  for(int i = 0; i < N; i++)
    I(i,i) = 1;

  SparseMatrix Laplacian_(N, N);
  cout << "Matrix size is: " << N << endl;

  SparseMatrix Laplacian1(N,N);
  cout << "Matrix product1" << endl;
  boost::numeric::ublas::axpy_prod(Dnim, A, Laplacian1);
  cout << "Matrix product2" << endl;
  SparseMatrix Laplacian2(N,N);
  boost::numeric::ublas::axpy_prod(Laplacian1, Dnim, Laplacian2);
  cout << "Matrix minus" << endl;
  Laplacian_ = I - Laplacian2;
  //cout << "Laplacian: " << endl << Laplacian_ << endl;

  cout << "Copying Laplacian_ to Laplacian" << endl;

  cv::SparseMat Laplacian(symA.dims(), symA.size(), symA.type());
  int dims[2];
  for(SparseMatrix::iterator1 it1 = Laplacian_.begin1(); it1 != Laplacian_.end1(); it1++){
    for(SparseMatrix::iterator2 it2 = it1.begin(); it2 != it1.end(); it2++){
      dims[0] = it2.index1();
      dims[1] = it2.index2();
      Laplacian.ref<float>(dims) = *it2;
    }
  }

  cout << "Returning" << endl;

  return Laplacian;
}

/****************************************************************************/
vector<int> normalizedSpectralClustering(const SparseMat& symA, const int k){
  if(symA.size()[0] != symA.size()[1]){
    cout << "Adjacency matrix is not symmetric..." << endl;
    exit(1);
  }

  int N = symA.size()[0];

  float myepsilon = 0.000000001;

  cout << "Getting Laplacian" << endl;

  SparseMat Laplacian = getNormalizedLaplacianMatrix(symA);
  
  cout << "Computing eigenvector" << endl;

  pair<Mat, Mat> eigenValVec = eigs(Laplacian, k, EIGS_SMALLEST);

  Mat eigenvalues(eigenValVec.first);
  Mat eigenvectors(eigenValVec.second);
  
  Mat eigCols = eigenvectors.t();
  
  float sum = 0;
  for(int i = 0; i < N; i++){
    sum = 0;
    for(int j = 0; j < k; j++)
      sum += pow(eigCols.at<float>(i,j), 2.0);
    sum = sqrt(sum) + myepsilon;
    for(int j = 0; j < k; j++)
      eigCols.at<float>(i,j) /= sum;
  }
  
  // clustering
  Mat idx(N, 1, CV_32S);
  for(int i = 0; i < N; i++)
    idx.at<int>(i,0) = -1;
  Mat centers(N, k, CV_32F);
  TermCriteria t(1, 100, myepsilon);
  kmeans(eigCols, k, idx, t, 5, KMEANS_PP_CENTERS, centers);

  vector<int> answer;
  for(int i = 0; i < N; i++)
    answer.push_back(idx.at<int>(i,0));

  for(int i = 0; i < answer.size(); i++)
    cout << answer[i] << " ";
  cout << endl;

  return answer;
}

/****************************************************************************/
/****************************************************************************/
vector<int> unnormalizedSpectralClustering(const SparseMat& symA, const int k){
  if(symA.size()[0] != symA.size()[1]){
    cout << "Adjacency matrix is not symmetric..." << endl;
    exit(1);
  }

  int N = symA.size()[0];

  float myepsilon = 0.000000001;

  cout << "Getting Laplacian" << endl;

  SparseMat Laplacian = getLaplacianMatrix<float>(symA);

  cout << "Eigenvector" << endl;

  pair<Mat, Mat> eigenValVec = eigs(Laplacian, k, EIGS_SMALLEST);

  Mat eigenvalues(eigenValVec.first);
  Mat eigenvectors(eigenValVec.second);
  
  Mat eigCols = eigenvectors.t();
  
  // clustering
  Mat idx(N, 1, CV_32S);
  for(int i = 0; i < N; i++)
    idx.at<int>(i,0) = -1;
  Mat centers(N, k, CV_32F);
  TermCriteria t(1, 100, myepsilon);
  kmeans(eigCols, k, idx, t, 5, KMEANS_PP_CENTERS, centers);

  vector<int> answer;
  for(int i = 0; i < N; i++)
    answer.push_back(idx.at<int>(i,0));

  for(int i = 0; i < answer.size(); i++)
    cout << answer[i] << " ";
  cout << endl;

  return answer;
}

/****************************************************************************/
/****************************************************************************/






