#ifndef SPARSE_STUFF_H
#define SPARSE_STUFF_H

#include <opencv2/opencv.hpp>
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/gpu/gpu.hpp>

/****************************************************************************/
enum {EIGS_SMALLEST, EIGS_LARGEST};

/* returns the few lowest/largest eigenvalues of matrix symA */
std::pair<cv::Mat, cv::Mat> eigs(const cv::SparseMat & symA, const int n_lowest_eigenval, const int type);

/****************************************************************************/
/* returns the normalized Laplacian matrix L = I - D^(-0.5) * A * D^(-0.5)*/
cv::SparseMat getNormalizedLaplacianMatrix(const cv::SparseMat& symA);

/****************************************************************************/
/* returns spectral clustering */
std::vector<int> normalizedSpectralClustering(const cv::SparseMat& A, const int k);
std::vector<int> unnormalizedSpectralClustering(const cv::SparseMat& A, const int k);
/****************************************************************************/
/* returns the diagonal matrix D = diag(sum(A)) */
template <class T>
cv::SparseMat matrixToD(const cv::SparseMat& matrix){
  int height = matrix.size()[0];
  int width = matrix.size()[1];
  cv::SparseMat D(matrix.dims(), matrix.size(), matrix.type());
  for(int i = 0; i < height; i++)
    D.ref<T>(i,i) = 0;

  int i,j;
  for(cv::SparseMat::const_iterator it = matrix.begin(); it != matrix.end(); it++){
    i = it.node()->idx[0];
    j = it.node()->idx[1];
    D.ref<T>(i,i) += it.value<T>();
  }

  return D;
}

/****************************************************************************/
/* returns the laplacian matrix L = D - A */
template <class T>
cv::SparseMat getLaplacianMatrix(const cv::SparseMat& A){
  if(A.size()[0] != A.size()[1]){
    std::cout << "getLaplacianMatrix: A is not a square matrix" << std::endl;
    exit(1);
  }
  int N = A.size()[0];
  cv::SparseMat D = matrixToD<T>(A);
  cv::SparseMat Laplacian(A.dims(), A.size(), A.type());
  int idx[2];
  for(int i = 0; i < N; i++){
    idx[0] = i;
    idx[1] = i;
    Laplacian.ref<T>(idx) = D.value<T>(idx) - A.value<T>(idx);
  }
  for(cv::SparseMat::const_iterator it = A.begin(); it != A.end(); it++){
    if(it.node()->idx[0] == it.node()->idx[1])
      continue;
    Laplacian.ref<T>(it.node()->idx) = D.value<T>(it.node()->idx) - it.value<T>();
  }
  return Laplacian;
}

/****************************************************************************/
template <class T>
void printMatrix(const cv::SparseMat& matrix){
  int i,j;
  for(cv::SparseMat::const_iterator it = matrix.begin(); it != matrix.end(); it++)
    std::cout << "(" << it.node()->idx[0] << "," << it.node()->idx[1] << "): " << it.value<T>() << std::endl;
  return;
}
/****************************************************************************/


#endif //SPARSE_STUFF_H
