#include "global.h"
#include "math.h"

#include <set>

using namespace std;
using namespace cv;

/************************************************************************************************************/
int timeToFrame(const int hour, const int minute, const int second, const int millisecond, const int fps){
  int total = (hour*3600 + minute*60 + second) * fps;
  total += (millisecond*fps)/1000;
  return total;
}

/************************************************************************************************************/
vector<int> randsample(int n, int k){
  vector<int> output;
  set<int> tempSet;
  while(output.size() < k){
    int r = rand() % n;
    if(tempSet.find(r) == tempSet.end()){
      output.push_back(r);
      tempSet.insert(r);
    }
  }
  return output;
}

/************************************************************************************************************/
Mat repmat(const Mat& M, int nRows, int nCols){
  Mat output(M.rows, M.cols, M.type());
  M.copyTo(output);
  
  if(nCols > 1){
    output = output.t();
    Mat temp(output.rows, output.cols, output.type());
    output.copyTo(temp);
    for(int i = 0; i < nCols-1; i++)
      output.push_back(temp);
    output = output.t();
  }
  
  if(nRows > 1){
    Mat temp(output.rows, output.cols, output.type());
    output.copyTo(temp);
    for(int i = 0; i < nRows-1; i++)
      output.push_back(temp);
  }

  return output;
}

/************************************************************************************************************/
void gradient(const Mat& M, Mat & returnMY, Mat & returnMX){
  int ysize = M.rows;
  int xsize = M.cols;
  
  returnMY = Mat::zeros(ysize, xsize, CV_32F);
  returnMX = Mat::zeros(ysize, xsize, CV_32F);

  Mat kx(1,3,CV_32F);
  kx.at<float>(0,0) = -0.5;
  kx.at<float>(0,2) =  0.5;

  Mat ky(3,1,CV_32F);
  ky.at<float>(0,0) = -0.5;
  ky.at<float>(2,0) =  0.5;

  filter2D(M, returnMX, -1, kx, Point(-1,-1), 0, BORDER_REPLICATE);
  filter2D(M, returnMY, -1, ky, Point(-1,-1), 0, BORDER_REPLICATE);
  for(int i = 0; i < ysize; i++){
    returnMX.at<float>(i,0) *= 2;
    returnMX.at<float>(i,xsize-1) *= 2;
  }
  for(int j = 0; j < xsize; j++){
    returnMY.at<float>(0,j) *= 2;
    returnMY.at<float>(ysize-1,j) *= 2;
  }

  return;
}

/************************************************************************************************************/
Mat powMatrix(const Mat& M, const float powerValue){
  int ysize = M.rows;
  int xsize = M.cols;
  Mat output(ysize, xsize, M.type());
  for(int i = 0; i < ysize; i++)
    for(int j = 0; j < xsize; j++)
      output.at<float>(i,j) = pow(M.at<float>(i,j), powerValue);

  return output;
}

/************************************************************************************************************/
Mat expMatrix(const Mat& M, const float powerValue){
  int ysize = M.rows;
  int xsize = M.cols;
  Mat output(ysize, xsize, M.type());
  for(int i = 0; i < ysize; i++)
    for(int j = 0; j < xsize; j++)
      output.at<float>(i,j) = exp(M.at<float>(i,j) * powerValue);

  return output;
}

/************************************************************************************************************/
float intersectHist(const std::vector<float>& hist1, const std::vector<float>& hist2){
  float output = 0;
  for(int i = 0; i < (int)(hist1.size()); i++)
    output += hist1[i] * hist2[i];
  return output;
}

/************************************************************************************************************/
double myUcharNorm(const uchar * v1, const uchar * v2, int length){
  double answer = 0;
  for(int i = 0; i < length; i++)
    answer += pow((double)(v1[i] - v2[i]), 2.0);
  answer = sqrt(answer);
  return answer;
}

/************************************************************************************************************/
Mat getRandomImage(const int height, const int width){
  Mat answer(height, width, CV_8UC3);
  typedef Vec<uchar,3> Vec3u;
  for(int i = 0; i < height; i++){
    for(int j = 0; j < width; j++){
      answer.at<Vec3u>(i,j)[0] = round(rand()*255);
      answer.at<Vec3u>(i,j)[1] = round(rand()*255);
      answer.at<Vec3u>(i,j)[2] = round(rand()*255);
    }
  }  
  return answer;
}

/************************************************************************************************************/
cv::Mat getLaplacianMatrix(const cv::Mat& adjacency){
  //  cout << "Computing the laplacian" << endl;
  // compute the laplacian
  /*  Mat I(Mat::eye(height, width, CV_32F));
  Mat D = matrixToD<float>(adjacency);
  Mat Dnim_(height, width, type);
  pow(D, -0.5, Dnim_);
  Mat Laplacian = Dnim_ * adjacency;
  Laplacian = Laplacian * Dnim_;
  Laplacian = I - Laplacian;*/
  Mat D = matrixToD<float>(adjacency);
  Mat Laplacian = D - adjacency;
  return Laplacian;
}

/************************************************************************************************************/
cv::Mat getNormalizedLaplacianMatrix(const cv::Mat& adjacency){
  //cout << "Computing the normalized laplacian" << endl;
  // compute the laplacian
  Size wh = adjacency.size();
  int height = wh.height;
  int width = wh.width;
  if(height != width){
    cout << "Adjacency matrix is not symmetric..." << endl;
    exit(0);
  }
  int type = adjacency.type();
  Mat I(Mat::eye(height, width, CV_32F));
  Mat D = matrixToD<float>(adjacency);
  Mat Dnim_(height, width, type);
  for(int i = 0; i < height; i++)
    for(int j = 0; j < width; j++)
      Dnim_.at<float>(i,j) = 0;
  for(int i = 0; i < height; i++)
    Dnim_.at<float>(i,i) = 1.0/sqrt(D.at<float>(i,i));
  Mat Laplacian1 = Dnim_ * adjacency;
  Mat Laplacian2 = Laplacian1 * Dnim_;
  Mat Laplacian = I - Laplacian2;
  return Laplacian;
}
/************************************************************************************************************/
vector<int> normalizedSpectralClustering(const cv::Mat& adjacency, const int k){
  Size wh = adjacency.size();
  int height = wh.height;
  int width = wh.width;
  if(height != width){
    cout << "Adjacency matrix is not symmetric..." << endl;
    exit(0);
  }
  int type = adjacency.type();

  double myepsilon = 0.000000000000001;

  Mat Laplacian = getNormalizedLaplacianMatrix(adjacency);

  double d = determinant(Laplacian);
  //  cout << "The determinant is: " << d << endl;
  if(d < myepsilon)
    cout << "The matrix is singular:((((((((((((((((((((((((" << endl;

  //cout << "Computing Eigenvectors" << endl;
  Mat eigenvectors(height, width, type);
  for(int i = 0; i < height; i++)
    for(int j = 0; j < width; j++)
      eigenvectors.at<float>(i,j) = -1;
  Mat eigenvalues(height, 1, type);
  for(int i = 0; i < height; i++)
    eigenvalues.at<float>(i,0) = -1;
  //eigen(Laplacian, eigenvalues, eigenvectors, height-k, height-1);
  eigen(Laplacian, eigenvalues, eigenvectors);
  //cout << "EigenValues:" << endl;
  for(int i = 0; i < height; i++)
    cout << eigenvalues.at<float>(i,0) << " ";
  cout << endl;
  //Mat eigRows = eigenvectors.rowRange(0, k);
  Mat eigRows = eigenvectors.rowRange(height-k, height);
  Mat eigCols = eigRows.t();
  print<float>(eigCols.rowRange(0,4), "eigCols");


  //cout << "Normalize the rows to 1" << endl;
  // normalize the rows to 1 
  for(int i = 0; i < width; i++){
    double sum = 0;
    for(int j = 0; j < k; j++)
      sum += pow((double)(eigCols.at<float>(i,j)), 2.0);
    sum = sqrt(sum) + myepsilon;
    for(int j = 0; j < k; j++)
      eigCols.at<float>(i,j) /= sum;
  }
  
  //cout << "Kmeans" << endl;
  // clustering
  Mat idx(width, 1, CV_32S);
  for(int i = 0; i < width; i++)
    idx.at<int>(i,0) = -1;
  Mat centers(width, k, type);
  TermCriteria t(1,100,myepsilon);
  kmeans(eigCols, k, idx, t, 5, KMEANS_PP_CENTERS, centers);

  vector<int> answer;
  for(int i = 0; i < width; i++)
    answer.push_back(idx.at<int>(i,0));

  for(int i = 0; i < answer.size(); i++)
    cout << answer[i] << " ";
  cout << endl;

  return answer;
}

/************************************************************************************************************/
/************************************************************************************************************/
vector<int> unnormalizedSpectralClustering(const cv::Mat& adjacency, const int k){
  Size wh = adjacency.size();
  int height = wh.height;
  int width = wh.width;
  if(height != width){
    cout << "Adjacency matrix is not symmetric..." << endl;
    exit(0);
  }
  int type = adjacency.type();

  double myepsilon = 0.000000000000001;

  Mat Laplacian = getLaplacianMatrix(adjacency);

  double d = determinant(Laplacian);
  //cout << "The determinant is: " << d << endl;
  if(d < myepsilon)
    cout << "The matrix is singular:((((((((((((((((((((((((" << endl;

  //cout << "Computing Eigenvectors" << endl;
  Mat eigenvectors(height, width, type);
  for(int i = 0; i < height; i++)
    for(int j = 0; j < width; j++)
      eigenvectors.at<float>(i,j) = -1;
  Mat eigenvalues(height, 1, type);
  for(int i = 0; i < height; i++)
    eigenvalues.at<float>(i,0) = -1;
  //eigen(Laplacian, eigenvalues, eigenvectors, height-k, height-1);
  eigen(Laplacian, eigenvalues, eigenvectors);
  //cout << "EigenValues:" << endl;
  for(int i = 0; i < height; i++)
    cout << eigenvalues.at<float>(i,0) << " ";
  cout << endl;
  //Mat eigRows = eigenvectors.rowRange(0, k);
  Mat eigRows = eigenvectors.rowRange(height-k-1, height-1);
  Mat eigCols = eigRows.t();
  print<float>(eigCols.rowRange(0,4), "eigCols");


  //cout << "Normalize the rows to 1" << endl;
  // normalize the rows to 1 
  /*  for(int i = 0; i < width; i++){
    double sum = 0;
    for(int j = 0; j < k; j++)
      sum += pow((double)(eigCols.at<float>(i,j)), 2.0);
    sum = sqrt(sum) + myepsilon;
    for(int j = 0; j < k; j++)
      eigCols.at<float>(i,j) /= sum;
      }*/
  
  //cout << "Kmeans" << endl;
  // clustering
  Mat idx(width, 1, CV_32S);
  for(int i = 0; i < width; i++)
    idx.at<int>(i,0) = -1;
  Mat centers(width, k, type);
  TermCriteria t(1,100,myepsilon);
  kmeans(eigCols, k, idx, t, 5, KMEANS_PP_CENTERS, centers);

  vector<int> answer;
  for(int i = 0; i < width; i++)
    answer.push_back(idx.at<int>(i,0));

  for(int i = 0; i < answer.size(); i++)
    cout << answer[i] << " ";
  cout << endl;

  return answer;
}

/************************************************************************************************************/
Mat maskOnImage(const Mat& mask, const Mat& image){
  typedef Vec<uchar,3> Vec3u;
  int h = image.rows;
  int w = image.cols;
  Mat newImage(h, w, CV_8UC3);

  if(mask.type() == CV_8UC1){
    for(int i = 0; i < h; i++){
      for(int j = 0; j < w; j++){
	if(mask.at<uchar>(i,j) == 0){
	  newImage.at<Vec3u>(i,j)[0] = 0;
	  newImage.at<Vec3u>(i,j)[1] = 0;
	  newImage.at<Vec3u>(i,j)[2] = 0;
	}
	else{
	  newImage.at<Vec3u>(i,j)[0] = image.at<Vec3u>(i,j)[0];
	  newImage.at<Vec3u>(i,j)[1] = image.at<Vec3u>(i,j)[1];
	  newImage.at<Vec3u>(i,j)[2] = image.at<Vec3u>(i,j)[2];
	}
      }
    }
  }

  return newImage;
}

/************************************************************************************************************/
Mat maskOnImage2(const Mat& mask, const Mat& image){
  typedef Vec<uchar,3> Vec3u;
  int h = image.rows;
  int w = image.cols;
  Mat newImage(h, w, CV_8UC3);

  if(mask.type() == CV_8UC1){
    for(int i = 0; i < h; i++){
      for(int j = 0; j < w; j++){
	if(mask.at<uchar>(i,j) == 0){
	  newImage.at<Vec3u>(i,j)[0] = min(image.at<Vec3u>(i,j)[0] + 200, 255);
	  newImage.at<Vec3u>(i,j)[1] = min(image.at<Vec3u>(i,j)[1] + 200, 255);
	  newImage.at<Vec3u>(i,j)[2] = min(image.at<Vec3u>(i,j)[2] + 200, 255);
	}
	else{
	  newImage.at<Vec3u>(i,j)[0] = image.at<Vec3u>(i,j)[0];
	  newImage.at<Vec3u>(i,j)[1] = image.at<Vec3u>(i,j)[1];
	  newImage.at<Vec3u>(i,j)[2] = image.at<Vec3u>(i,j)[2];
	}
      }
    }
  }

  return newImage;
}

/************************************************************************************************************/
/************************************************************************************************************/
