#ifndef GLOBAL_H
#define GLOBAL_H

#include <opencv2/opencv.hpp>
#include <vector>

typedef unsigned char uchar;
typedef unsigned int uint;
typedef unsigned short ushort;

//const int alireza = 75;
//typedef sfcnn_knng<unsigned char*, alireza, unsigned char> sfcnn_knng5;
//typedef sfcnn_knng<unsigned char*, 243, unsigned char> sfcnn_knng9;

/* returns k random numbers between 0 and n-1 */
std::vector<int> randsample(int n, int k);

float intersectHist(const std::vector<float>& hist1, const std::vector<float>& hist2);

int timeToFrame(const int hour, const int minute, const int second, const int millisecond, const int fps);

/* given to vectors of uchar, it returns the norm of their differences */
double myUcharNorm(const uchar * v1, const uchar * v2, int length);

/* Compute Laplacian */
cv::Mat getLaplacianMatrix(const cv::Mat& adjacency);
cv::Mat getNormalizedLaplacianMatrix(const cv::Mat& adjacency);

/* given an adjacency matrix, and number of cluster k, returns a vector of cluster numbers assigned to each element */
std::vector<int> normalizedSpectralClustering(const cv::Mat& adjacency, const int k);
std::vector<int> unnormalizedSpectralClustering(const cv::Mat& adjacency, const int k);

/* returns an image with random RGB values for its pixels */
cv::Mat getRandomImage(const int height, const int width);

/* repmat */
cv::Mat repmat(const cv::Mat& M, int nRows, int nCols);

cv::Mat maskOnImage(const cv::Mat& mask, const cv::Mat& image);
cv::Mat maskOnImage2(const cv::Mat& mask, const cv::Mat& image);

/* prints a matrix */
template<class T>
void print(const cv::Mat& matrix, char * name){
  if(matrix.channels() > 1)
    return;
  cv::Size wh = matrix.size();
  int height = wh.height;
  int width = wh.width;
  if(width > 20)
    return;
  std::cout << name << std::endl;
  std::cout << " \t";
  for(int j = 0; j < width; j++)
    std::cout << j+1 << "\t";
  std::cout << std::endl;
  for(int j = 0; j < 150; j++)
    std::cout << "-";
  std::cout << std::endl;
  for(int i = 0; i < height; i++){
    std::cout << i+1 << "|\t";
    for(int j = 0; j < width; j++){
      std::cout << matrix.at<T>(i,j) << "\t";
    }
    std::cout << std::endl;
  }
  std::cout << std::endl;
  
  return;
}

/* returns the diagonal sum of the matrix, used for laplacian */
template<class T>
cv::Mat matrixToD(const cv::Mat& matrix){
  int channels = matrix.channels();
  if(channels != 1){
    std::cout << "matrixToD: the matrix should have only one channel!" << std::endl;
    exit(0);
  }
  cv::Size wh = matrix.size();
  int height = wh.height;
  int width = wh.width;
  cv::Mat answer(height, width, matrix.type());
  for(int i = 0; i < height; i++)
    for(int j = 0; j < width; j++)
      answer.at<T>(i,j) = 0;
  for(int i = 0; i < height; i++){
    T temp = 0;
    for(int j = 0; j < width; j++)
      temp += matrix.at<T>(i,j);
    answer.at<T>(i,i) = temp;
  }

  return answer;
}

/* returns gradients in y and x directions */
void gradient(const cv::Mat& M, cv::Mat & returnMY, cv::Mat & returnMX);

/* raises matrix M to pow */
cv::Mat powMatrix(const cv::Mat& M, const float powerValue);

/* changes M(i,j) to exp(M(i,j)*powerValue) */
cv::Mat expMatrix(const cv::Mat& M, const float powerValue);

/* subsamples the matrix at every st steps */
template<class T>
cv::Mat subMatrix(const cv::Mat& matrix, const int st){
  int height = matrix.rows;
  int width = matrix.cols;
  int type = matrix.type();
  int height2 = ((height-1)/st) + 1;
  int width2 = ((width-1)/st) + 1;
  cv::Mat output(height2, width2, type);
  for(int i = 0; i < height2; i++)
    for(int j = 0; j < width2; j++)
      output.at<T>(i,j) = matrix.at<T>(i*st, j*st);

  return output;
}

/* computes the submatrix, for the locations that the vector values are non-zero 
   matrix is mxn
   prRows is mx1, CV_32S
   prCols is nx1, CV_32S
*/
template<class T>
cv::Mat subMatrix(const cv::Mat& matrix, const cv::Mat& prRows, const cv::Mat& prCols){
  int height = matrix.rows;
  int width = matrix.cols;
  int type = matrix.type();

  if(height != prRows.rows){
    std::cout << "Number of elements in prRows doesn't match the height of the matrix" << std::endl;
    exit(1);
  }

  if(width != prCols.rows){
    std::cout << "Number of elements in prCols doesn't match the width of the matrix" << std::endl;
    exit(1);
  }

  int numR = 0; // number of non-zero elements in prRows
  int numC = 0; // number of non-zero elements in prCols
  for(int i = 0; i < height; i++)
    if(prRows.at<int>(i,0) != 0)
      numR++;
  for(int i = 0; i < width; i++)
    if(prCols.at<int>(i,0) != 0)
      numC++;

  cv::Mat Mra(numR, width, type);
  int countR = -1;
  for(int i = 0; i < height; i++){
    if(prRows.at<int>(i,0) == 0)
      continue;
    countR++;
    for(int j = 0; j < width; j++)
      Mra.at<T>(countR,j) = matrix.at<T>(i,j);
  }

  cv::Mat Mrc(numR, numC, type);
  int countC = -1;
  for(int j = 0; j < width; j++){
    if(prCols.at<int>(j,0) == 0)
      continue;
    countC++;
    for(int i = 0; i < numR; i++)
      Mrc.at<T>(i,countC) = Mra.at<T>(i,j);
  }

  return Mrc;
}

/* returns only the selected rows in indices */
template<class T>
cv::Mat indMatrixRows(const cv::Mat& M, const std::vector<int>& indices){
  int height = M.rows;
  int width = M.cols;
  
  cv::Mat output((int)(indices.size()), width, M.type());

  int count = 0;
  for(std::vector<int>::const_iterator it = indices.begin(); it != indices.end(); it++){
    int i = *it;
    for(int j = 0; j < width; j++)
      output.at<T>(count, j) = M.at<T>(i,j);
    count++;
  }

  return output;
}

/* returns only the selected columns in indices */
template<class T>
cv::Mat indMatrixCols(const cv::Mat& M, const std::vector<int>& indices){
  int height = M.rows;
  int width = M.cols;
  
  cv::Mat output(height, (int)(indices.size()), M.type());

  int count = 0;
  for(std::vector<int>::const_iterator it = indices.begin(); it != indices.end(); it++){
    int j = *it;
    for(int i = 0; i < height; i++)
      output.at<T>(i, count) = M.at<T>(i,j);
    count++;
  }

  return output;
}


/* returns the maximum value of the matrix */
template<class T>
T matrixMaxValue(const cv::Mat& inputMatrix, int& maxY, int& maxX){
  cv::Size wh = inputMatrix.size();
  int width = wh.width;
  int height = wh.height;
  T answer = -100000;
  int count = 0;
  maxY = -1;
  maxX = -1;
  for(int i = 0; i < height; i++){
    for(int j = 0; j < width; j++){
      T temp = inputMatrix.at<T>(i,j);
      if(temp >= answer){
	answer = temp;
	maxY = i;
	maxX = j;
      }
    }
  }
  return answer;
}

/* returns the maximum value of the diagonal elements of the matrix */
template<class T>
T matrixMaxValueDiagonal(const cv::Mat& inputMatrix, int& maxY, int& maxX){
  cv::Size wh = inputMatrix.size();
  int width = wh.width;
  int height = wh.height;
  T answer = -10000000;
  int count = 0;
  maxY = -1;
  maxX = -1;
  for(int i = 0; i < height; i++){
    T temp = inputMatrix.at<T>(i,i);
    if(temp >= answer){
      answer = temp;
      maxY = i;
      maxX = i;
    }
  }
  return answer;
}

/* returns the minimum value of the matrix */
template<class T>
T matrixMinValue(const cv::Mat& inputMatrix, int& minY, int& minX){
  cv::Size wh = inputMatrix.size();
  int width = wh.width;
  int height = wh.height;
  T answer = 10000000;
  int count = 0;
  minY = -1;
  minX = -1;
  for(int i = 0; i < height; i++){
    for(int j = 0; j < width; j++){
      T temp = inputMatrix.at<T>(i,j);
      if(temp <= answer){
	answer = temp;
	minY = i;
	minX = j;
      }
    }
  }
  return answer;
}

/* returns the minimum value of the matrix which is greater than zero */
template<class T>
T matrixMinValueGreaterThanZero(const cv::Mat& inputMatrix, int& minY, int& minX){
  cv::Size wh = inputMatrix.size();
  int width = wh.width;
  int height = wh.height;
  T answer = 10000000;
  int count = 0;
  minY = -1;
  minX = -1;
  for(int i = 0; i < height; i++){
    for(int j = 0; j < width; j++){
      T temp = inputMatrix.at<T>(i,j);
      if(temp <= 0)
	continue;
      if(temp <= answer){
	answer = temp;
	minY = i;
	minX = j;
      }
    }
  }
  return answer;
}

/* returns the minimum value of the diagonal elements of the matrix */
template<class T>
T matrixMinValueDiagonal(const cv::Mat& inputMatrix, int& minY, int& minX){
  cv::Size wh = inputMatrix.size();
  int width = wh.width;
  int height = wh.height;
  T answer = 10000000;
  int count = 0;
  minY = -1;
  minX = -1;
  for(int i = 0; i < height; i++){
    T temp = inputMatrix.at<T>(i,i);
    if(temp <= answer){
      answer = temp;
      minY = i;
      minX = i;
    }
  }
  return answer;
}

/* given two sparse matrices merges them, as it sets the value of each element from the two matrices */
template<class T>
cv::SparseMat mergeSparseMatrices(const cv::SparseMat& s1, const cv::SparseMat& s2){
  int dims1 = s1.dims();
  int dims2 = s2.dims();
  if(dims1 != dims2){
    std::cout << "mergeSparseMatrices: matrice dimensions do not match" << std::endl;
    exit(0);
  }
  const int * sizes1 = s1.size();
  const int * sizes2 = s2.size();
  for(int i = 0; i < dims1; i++){
    if(sizes1[i] != sizes2[i]){
      std::cout << "mergeSparseMatrices: sizes do not match" << std::endl;
      exit(0);
    }
  }
  int type1 = s1.type();
  int type2 = s2.type();
  if(type1 != type2){
    std::cout << "mergeSparseMatrices: types do not match" << std::endl;
    exit(0);
  }
  
  cv::SparseMat answer(dims1, sizes1, type1);

  for(cv::SparseMat::const_iterator it = s1.begin(); it != s1.end(); ++it){
    answer.ref<T>(it.node()->idx) = it.value<T>();
  }

  for(cv::SparseMat::const_iterator it = s2.begin(); it != s2.end(); ++it){
    answer.ref<T>(it.node()->idx) = it.value<T>();
  }

  return answer;
}


#endif
